from PyQt5.QtCore import QPropertyAnimation
from PyQt5.QtWidgets import (QWidget, QTabBar, QSizePolicy, QStackedWidget,
                             QHBoxLayout)


class Toolbox(QWidget):
    """A side-oriented widget similar to a TabWidget that can be collapsed and
    expanded.

    A Toolbox is designed to be a container for sets of controls, grouped into
    'pages' and accessible by a TabBar, in the same way as a TabWidget.
    A page is normally a QWidget with a layout that contains controls.
    A widget can be added as a new tab using :meth:`addTab`.
    The Toolbox has slots for triggering its collapse and expansion, both in an
    animated mode (soft slide) and a 'quick' mode which skips the animation.
    Commonly the collapse/expand slots are connected to the tabBar's
    :meth:`tabBarDoubleClicked` signal. Normally in the DataLogger a Toolbox is
    created and then added to a :class:`~cued_datalogger.api.toolbox.MasterToolbox`,
    which connects the relevant signals for collapsing and expanding the
    Toolbox.


    Attributes
    ----------
    tabBar : QTabBar
    tabPages : QStackedWidget
        The stack of widgets that form the pages of the tabs.
    collapse_animation : QPropertyAnimation
        The animation that controls how the Toolbox collapses.
    """

    def __init__(self, widget_side='left', parent=None):
        self.parent = parent

        super().__init__(parent)

        layout = QHBoxLayout(self)
        layout.setSpacing(0)

        # # Create the tab bar
        self.tabBar = QTabBar(self)

        self.tabBar.setTabsClosable(False)
        self.tabBar.setMovable(False)
        self.tabBar.setSizePolicy(QSizePolicy.Expanding,
                                  QSizePolicy.Expanding)
        # # Create the Stacked widget for the pages
        self.tabPages = QStackedWidget(self)
        self.tabPages.setSizePolicy(QSizePolicy.Expanding,
                                    QSizePolicy.Expanding)

        # # Link the signals so that changing tab leads to a change of page
        self.tabBar.currentChanged.connect(self.changePage)

        # # Add them to the splitter (self)
        # Right side orientation
        if widget_side == 'right':
            self.tabBar.setShape(QTabBar.RoundedWest)

            layout.addWidget(self.tabBar)
            layout.addWidget(self.tabPages)

        # Left side orientation
        else:
            self.tabBar.setShape(QTabBar.RoundedEast)

            layout.addWidget(self.tabPages)
            layout.addWidget(self.tabBar)

        # # Create the animation
        self.collapse_animation = QPropertyAnimation(self.tabPages,
                                                     b'maximumWidth')
        self.collapse_animation.setDuration(250)
        self.collapse_animation.setEndValue(0)
        self.collapse_animation.finished.connect(self.on_animation_finished)

        self.collapsed = False
        self.max_width = 250

    def addTab(self, widget, title):
        """Add a new tab, with the page widget *widget* and tab title
        *title*."""
        self.tabBar.addTab(title)
        self.tabPages.addWidget(widget)

    def removeTab(self, title):
        """Remove the tab with title *title*."""
        for tab_num in range(self.tabBar.count()):
            if self.tabBar.tabText(tab_num) == title:
                self.tabBar.removeTab(tab_num)
                self.tabPages.removeWidget(self.tabPages.widget(tab_num))

    def toggle_collapse(self):
        """If collapsed, expand the widget so the pages are visible. If not
        collapsed, collapse the widget so that only the tabBar is showing."""
        # If collapsed, expand
        if self.collapsed:
            self.expand()
        # If expanded, collapse:
        else:
            self.collapse()

    def on_animation_finished(self):
        if self.collapsed:
            self.tabPages.hide()

    def expand(self):
        """Expand the widget so that the pages are visible."""
        self.collapse_animation.setStartValue(self.max_width)

        self.tabPages.show()
        self.collapse_animation.setDirection(QPropertyAnimation.Backward)
        self.collapse_animation.start()
        self.collapsed = False

    def collapse(self):
        """Collapse the widget so that only the tab bar is visible."""
        self.collapse_animation.setStartValue(self.max_width)

        self.collapse_animation.setDirection(QPropertyAnimation.Forward)
        self.collapse_animation.start()
        self.collapsed = True

    def fast_collapse(self):
        """Collapse the widget without the animation."""
        self.tabPages.hide()
        self.tabPages.setMaximumWidth(0)
        self.collapsed = True

    def fast_expand(self):
        """Expand the widget without the animation."""
        self.tabPages.show()
        self.tabPages.setMaximumWidth(self.max_width)
        self.collapsed = False

    def changePage(self, index):
        """Set the current page to *index*."""
        self.tabBar.setCurrentIndex(index)
        self.tabPages.setCurrentIndex(index)

        self.tabPages.setMaximumWidth(self.max_width)

        if self.tabPages.currentWidget():
            self.tabPages.currentWidget().resize(self.tabPages.size())

    def clear(self):
        """Remove all tabs and pages."""
        for i in range(self.tabBar.count()):
            # Remove the tab and page at position 0
            self.tabBar.removeTab(0)
            self.tabPages.removeWidget(self.tabPages.currentWidget())


class MasterToolbox(QStackedWidget):
    """A QStackedWidget of one or more Toolboxes that toggle collapse when
    the tabBar is double clicked.

    In the MasterToolbox, only the top Toolbox is expanded, and all the others
    are collapsed. When the index is changed with :meth:`set_toolbox`, the
    top Toolbox is changed and all other Toolboxes are collapsed and hidden.
    The MasterToolbox is the normal location for all tools and controls in the
    DataLogger.

    Attributes
    ----------
    Inherited attributes :
        See ``PyQt5.QtWidgets.QStackedWidget`` for inherited attributes.
    """
    def __init__(self, parent=None):
        super().__init__(parent)

    def toggle_collapse(self):
        """Toggle collapse of the MasterToolbox by toggling the collapse of the
        Toolbox that is on top."""
        self.currentWidget().toggle_collapse()

    def set_toolbox(self, toolbox_index):
        """Set current Toolbox to the Toolbox given by *toolbox_index*, by
        quick-collapsing and hiding all of the other Toolboxes. The new
        current Toolbox will be in the same collapse/expand state as the
        former current Toolbox (ie if the previous Toolbox was collapsed,
        the new current Toolbox will be collapsed, and vice versa)."""
        # Save the old toolbox
        old_toolbox_collapsed = self.currentWidget().collapsed

        # Collapse and hide all other toolboxes
        for i in range(self.count()):
            if i == toolbox_index:
                continue
            else:
                self.widget(i).fast_collapse()
                self.widget(i).hide()

        # Collapse / Expand the toolbox
        if old_toolbox_collapsed:
            self.widget(toolbox_index).fast_collapse()
        else:
            self.widget(toolbox_index).fast_expand()

        # Set the current toolbox
        self.setCurrentIndex(toolbox_index)

    def add_toolbox(self, toolbox):
        """Add a Toolbox to the MasterToolbox stack."""
        # Set the Toolbox's parent
        toolbox.parent = self
        toolbox.setParent(self)

        # When the Toolbox's tabBar is clicked, collapse the Toolbox
        toolbox.tabBar.tabBarDoubleClicked.connect(self.toggle_collapse)

        # Add the Toolbox to the stack
        self.addWidget(toolbox)



import sys,traceback

from cued_datalogger.api.numpy_extensions import to_dB
from cued_datalogger.api.pyqt_widgets import BaseNControl, MatplotlibCanvas
from cued_datalogger.api.pyqtgraph_extensions import ColorMapPlotWidget
from cued_datalogger.api.toolbox import Toolbox

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QSlider, QPushButton, QLabel, QSpinBox, QHBoxLayout, QGridLayout

import numpy as np

import scipy.signal


class MatplotlibSonogramContourWidget(MatplotlibCanvas):
    """A MatplotlibCanvas widget displaying the Sonogram contour plot."""

    def __init__(self, sonogram_toolbox=None,
                 channel=None,
                 contour_spacing_dB=None,
                 num_contours=None):
        self.sonogram_toolbox = sonogram_toolbox
        self.channel = channel
        self.contour_spacing_dB = contour_spacing_dB
        self.num_contours = num_contours

        #self.sonogram_toolbox.num_contours_slider.valueChanged.connect(self.update_plot)
        #self.sonogram_toolbox.num_contours_spinbox.valueChanged.connect(self.update_plot)
        #self.sonogram_toolbox.contour_spacing_slider.valueChanged.connect(self.update_plot)
        #self.sonogram_toolbox.contour_spacing_spinbox.valueChanged.connect(self.update_plot)

        MatplotlibCanvas.__init__(self, "Sonogram: Contour Plot")

        self.update_plot()

    def update_plot(self):
        """Redraw the sonogram on the canvas."""
        if self.channel is not None:
            self.F_bins, self.T_bins = np.meshgrid(self.channel.data("sonogram_frequency"),
                                                   self.channel.data("sonogram_time"))

            self.axes.clear()

            self.update_contour_sequence()

            self.axes.contour(self.F_bins, self.T_bins,
                              to_dB(np.abs(self.channel.data("sonogram"))),
                              self.contour_sequence)

            self.axes.set_xlabel('Freq (Hz)')
            self.axes.set_ylabel('Time (s)')

            self.axes.set_xlim(self.channel.data("sonogram_frequency").min(),
                               self.channel.data("sonogram_frequency").max())
            self.axes.set_ylim(self.channel.data("sonogram_time").min(),
                               self.channel.data("sonogram_time").max())

            self.draw()

    def update_contour_sequence(self):
        """Update the array which says where to plot contours, how many etc."""
        if self.channel is not None:
            # Create a vector with the right spacing from min to max value
            self.contour_sequence = np.arange(to_dB(np.abs(self.channel.data("sonogram"))).min(),
                                              to_dB(np.abs(self.channel.data("sonogram"))).max(),
                                              self.contour_spacing_dB)
            # Take the appropriate number of contours
            self.contour_sequence = self.contour_sequence[-self.num_contours:]

    def update_contour_spacing(self, value):
        """Slot for updating the plot when the contour spacing is changed."""
        self.contour_spacing_dB = value
        self.update_plot()

    def update_num_contours(self, value):
        """Slot for updating the plot when the number of contours is changed."""
        self.num_contours = value
        self.update_plot()

    def set_selected_channels(self, selected_channels):
        """Update which channel is being plotted."""
        # If no channel list is given
        if not selected_channels:
            self.channel = None
        else:
            self.channel = selected_channels[0]
            self.update_plot()


class SonogramDisplayWidget(ColorMapPlotWidget):
    """
    The SonogramDisplayWidget is the main display widget for everything in
    the sonogram domain.
    """
    def __init__(self, parent=None,
                 window_width=256,
                 window_overlap_fraction=8,
                 contour_spacing_dB=5,
                 num_contours=5):

        super().__init__(parent)
        self.parent = parent

        self.window_width = window_width
        self.window_overlap_fraction = window_overlap_fraction
        self.contour_spacing_dB = contour_spacing_dB
        self.num_contours = num_contours

        self.PlotWidget.setLabel('bottom', "Frequency", "Hz")
        self.PlotWidget.setLabel('left', "Time", "s")

        self.show()

    def update_window_width(self, value):
        """Slot for updating the plot when the window width is changed."""
        self.window_width = value
        self.update_plot()

    def update_window_overlap_fraction(self, value):
        """Slot for updating the plot when the window overlap fraction is changed."""
        self.window_overlap_fraction = value
        self.update_plot()

    def update_contour_spacing(self, value):
        """Slot for updating the plot when the contour spacing is changed."""
        self.contour_spacing_dB = value
        self.update_plot()

    def update_num_contours(self, value):
        """Slot for updating the plot when the number of contours is changed."""
        self.num_contours = value
        self.update_plot()

    def calculate_sonogram(self):
        """Calculate the sonogram, and store the values in the channel
        (including autogenerated datasets). Sonogram data is in complex form."""

        (self.freqs,
         self.times,
         self.FT) = scipy.signal.spectrogram(self.channel.data("time_series"),
                                           self.channel.metadata("sample_rate"),
                                           window=scipy.signal.get_window('hann', self.window_width),
                                           nperseg=self.window_width,
                                           noverlap=self.window_width // self.window_overlap_fraction,
                                           return_onesided=False,
                                           mode = 'complex')
        print(self.FT.shape)
        # SciPy's spectrogram gives the FT transposed, so we need to transpose it back
        self.FT = self.FT.transpose()
        # Scipy calculates all the conjugate spectra/frequencies as well -
        # we only want the positive ones
        self.freqs = np.abs(self.freqs[:self.freqs.size // 2 + 1])
        self.FT = self.FT[:, :self.FT.shape[1] // 2 + 1]

        self.channel.add_dataset("sonogram_frequency", data=self.freqs, units="Hz")
        self.channel.add_dataset("sonogram_omega", data=self.freqs*2*np.pi, units="rad")
        self.channel.add_dataset("sonogram_time", data=self.times, units="s")
        #self.channel.add_dataset("sonogram", data=self.FT, units=None)
        self.channel.add_dataset("sonogram", data=self.FT, units=None)

        # For backwards compatibility (and probably decay rate calculation in the future)
        self.channel.add_dataset("sonogram_phase", data=np.angle(self.FT), units=None)
        self.channel.add_dataset("sonogram_step", data=self.window_width // self.window_overlap_fraction, units=None)

    def update_plot(self):
        """Recalculate the sonogram, clear the canvas and replot."""
        if self.channel is not None:
            if self.channel.is_dataset("time_series"):
                self.calculate_sonogram()
                self.clear()
                self.plot_colormap(self.channel.data("sonogram_frequency"),
                                   self.channel.data("sonogram_time"),
                                   to_dB(np.abs(self.channel.data("sonogram"))),
                                   num_contours=self.num_contours,
                                   contour_spacing_dB=self.contour_spacing_dB)
        else:
            self.clear()

    def set_selected_channels(self, selected_channels):
        """Update which channel is being plotted."""
        # If no channel list is given
        if not selected_channels:
            self.channel = None
        else:
            self.channel = selected_channels[0]
            self.update_plot()


class SonogramToolbox(Toolbox):
    """Toolbox containing Sonogram controls."""

    sig_window_width_changed = pyqtSignal(int)
    sig_window_overlap_fraction_changed = pyqtSignal(int)
    sig_num_contours_changed = pyqtSignal(int)
    sig_contour_spacing_changed = pyqtSignal(int)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.parent = parent

        self.window_width = 256
        self.window_overlap_fraction = 8
        self.num_contours = 5
        self.contour_spacing_dB = 5

        self.init_ui()


    def init_ui(self):
        #------------Window width controls------------
        self.window_width_label = QLabel(self)
        self.window_width_label.setText("Window width")
        # Create control
        self.window_width_control = BaseNControl(Qt.Vertical, self)
        self.window_width_control.set_power_range(0, 10)
        self.window_width_control.set_value(self.window_width)
        self.window_width_control.valueChanged.connect(self.sig_window_width_changed)

        #------------Window increment controls------------
        self.window_overlap_fraction_label = QLabel(self)
        self.window_overlap_fraction_label.setText("Window overlap fraction")
        # Create control
        self.window_overlap_fraction_control = BaseNControl(Qt.Vertical, self)
        self.window_overlap_fraction_control.set_power_range(0, 6)
        self.window_overlap_fraction_control.set_value(self.window_overlap_fraction)
        self.window_overlap_fraction_control.valueChanged.connect(self.sig_window_overlap_fraction_changed.emit)

        #------------Contour spacing controls------------
        self.contour_spacing_label = QLabel(self)
        self.contour_spacing_label.setText("Contour spacing")
        # Create spinbox
        self.contour_spacing_spinbox = QSpinBox(self)
        self.contour_spacing_spinbox.setObjectName("contour_spacing_spinbox")
        self.contour_spacing_spinbox.setRange(1, 12)
        # Create slider
        self.contour_spacing_slider = QSlider(Qt.Vertical, self)
        self.contour_spacing_slider.setObjectName("contour_spacing_slider")
        self.contour_spacing_slider.setRange(1, 12)
        # Connect spinbox and slider together
        self.contour_spacing_spinbox.valueChanged.connect(self.contour_spacing_slider.setValue)
        self.contour_spacing_slider.valueChanged.connect(self.contour_spacing_spinbox.setValue)
        # Set values
        self.contour_spacing_spinbox.setValue(self.contour_spacing_dB)
        self.contour_spacing_slider.setValue(self.contour_spacing_dB)
        # Update screen on change
        self.contour_spacing_slider.valueChanged.connect(self.sig_contour_spacing_changed.emit)
        self.contour_spacing_spinbox.valueChanged.connect(self.sig_contour_spacing_changed.emit)

        #------------Num contours controls------------
        self.num_contours_label = QLabel(self)
        self.num_contours_label.setText("Num contours")
        # Create spinbox
        self.num_contours_spinbox = QSpinBox(self)
        self.num_contours_spinbox.setObjectName("num_contours_spinbox")
        self.num_contours_spinbox.setRange(1, 12)
        # Create slider
        self.num_contours_slider = QSlider(Qt.Vertical, self)
        self.num_contours_slider.setObjectName("num_contours_slider")
        self.num_contours_slider.setRange(1, 12)
        # Connect spinbox and slider together
        self.num_contours_spinbox.valueChanged.connect(self.num_contours_slider.setValue)
        self.num_contours_slider.valueChanged.connect(self.num_contours_spinbox.setValue)
        # Set values
        self.num_contours_spinbox.setValue(self.num_contours)
        self.num_contours_slider.setValue(self.num_contours)
        # Update screen on change
        self.num_contours_slider.valueChanged.connect(self.sig_num_contours_changed.emit)
        self.num_contours_spinbox.valueChanged.connect(self.sig_num_contours_changed.emit)

        #------------Matplotlib window controls---------
        # Create button
        self.convert_to_contour_btn = QPushButton("Show as contour plot", self)
        self.convert_to_contour_btn.resize(self.convert_to_contour_btn.sizeHint())
        self.convert_to_contour_btn.clicked.connect(self.open_contour_plot)

        #------------Layout------------
        # Sonogram controls:
        self.sonogram_controls_tab = QWidget(self)

        sonogram_controls_layout = QGridLayout()
        sonogram_controls_layout.addWidget(self.window_width_label, 0, 0)
        sonogram_controls_layout.addWidget(self.window_width_control, 1, 0)
        sonogram_controls_layout.addWidget(self.window_overlap_fraction_label, 0, 1)
        sonogram_controls_layout.addWidget(self.window_overlap_fraction_control, 1, 1)

        self.sonogram_controls_tab.setLayout(sonogram_controls_layout)

        # Plot controls:
        self.plot_controls_tab = QWidget(self)

        plot_controls_layout = QGridLayout()
        plot_controls_layout.addWidget(self.contour_spacing_label, 1, 0)
        plot_controls_layout.addWidget(self.contour_spacing_spinbox, 2, 0)
        plot_controls_layout.addWidget(self.contour_spacing_slider, 3, 0)
        plot_controls_layout.addWidget(self.num_contours_label, 1, 1)
        plot_controls_layout.addWidget(self.num_contours_spinbox, 2, 1)
        plot_controls_layout.addWidget(self.num_contours_slider, 3, 1)

        self.plot_controls_tab.setLayout(plot_controls_layout)

        # Export:
        self.export_tab = QWidget(self)

        export_layout = QGridLayout()
        export_layout.addWidget(self.convert_to_contour_btn, 0, 0)
        export_layout.setRowStretch(1,1)

        self.export_tab.setLayout(export_layout)

        #-------------Add tabs-----------------
        self.addTab(self.plot_controls_tab, "Plot Controls")
        self.addTab(self.sonogram_controls_tab, "Sonogram Controls")
        self.addTab(self.export_tab, "Export")

    def open_contour_plot(self):
        if hasattr(self, 'contour_plot'):
            self.contour_plot.close()
            delattr(self, 'contour_plot')
        else:
            self.contour_plot = MatplotlibSonogramContourWidget(channel=self.channel,
                                                                contour_spacing_dB=self.contour_spacing_dB,
                                                                num_contours=self.num_contours)
            self.sig_contour_spacing_changed.connect(self.contour_plot.update_contour_spacing)
            self.sig_num_contours_changed.connect(self.contour_plot.update_num_contours)
            self.contour_plot.show()

    def set_selected_channels(self, selected_channels):
        """Update which channel is being plotted"""
        # If no channel list is given
        if not selected_channels:
            self.channel = None
        else:
            self.channel = selected_channels[0]

        if hasattr(self, 'contour_plot'):
            self.contour_plot.set_selected_channels(selected_channels)

def func_1(t, w, x, A=4e3):
    """A simple decaying sine wave function."""
    return A * np.exp((1j*w - x)*t)


def function_generator(t):
    """A simple function generator with echoes."""
    f1 = func_1(t, 2000*2*np.pi, 2)
    f2 = func_1(t, 500*2*np.pi, 1)
    # Create an echo of one of the functions
    f1[f1.size//2:] += f1[:f1.size//2]
    result = f1 + f2
    return result


if __name__ == '__main__':
    duration = 10.0
    t = np.arange(0.0, duration, 1/4096)
    sig = function_generator(t)

    app = 0

    app = QApplication(sys.argv)

    w = QWidget()

    hbox = QHBoxLayout()
    w.setLayout(hbox)

    toolbox = SonogramToolbox(w)
    displaywidget = SonogramDisplayWidget()
    displaywidget.update_plot(sig)

    hbox.addWidget(toolbox)
    hbox.addWidget(displaywidget)

    toolbox.contour_spacing_slider.valueChanged.connect(displaywidget.update_contour_spacing)
    toolbox.contour_spacing_spinbox.valueChanged.connect(displaywidget.update_contour_spacing)

    toolbox.num_contours_slider.valueChanged.connect(displaywidget.update_num_contours)
    toolbox.num_contours_spinbox.valueChanged.connect(displaywidget.update_num_contours)

    toolbox.window_overlap_fraction_control.valueChanged.connect(displaywidget.update_window_overlap_fraction)

    toolbox.window_width_control.valueChanged.connect(displaywidget.update_window_width)

    w.show()

    sys.exit(app.exec_())




import sys, traceback
if __name__ == '__main__':
    sys.path.append('../../')

from cued_datalogger.api.channel import ChannelSet
from cued_datalogger.api.workspace import Workspace
from cued_datalogger.api.file_import import import_from_mat
from cued_datalogger.api.toolbox import Toolbox
from cued_datalogger.api.numpy_extensions import from_dB, to_dB, sdof_modal_peak
from cued_datalogger.api.pyqtgraph_extensions import InteractivePlotWidget

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5 import QtGui
from PyQt5.QtWidgets import (QApplication, QWidget, QGridLayout, QTableWidget,
                             QDoubleSpinBox, QCheckBox, QPushButton, QGroupBox,
                             QVBoxLayout, QHBoxLayout, QLabel, QComboBox,
                             QFileDialog, QTreeWidget, QTreeWidgetItem, QRadioButton)

import numpy as np
from scipy.optimize import leastsq, minimize

from pyqtgraph.Qt import QtCore
import pyqtgraph as pg
defaultpen='k'


def fit_circle_to_data(x, y):
    """
    Fit a geometric circle to the data given in x, y.

    Parameters
    ----------
    x : ndarray
    y : ndarray

    Returns
    -------
    x0 : float
        The x-coordinate of the centre of the circle.
    y0 : float
        The y-coordinate of the centre of the circle.
    R0 : float
        The radius of the circle.

    Notes
    -----
    This function solves a standard eigenvector formulation of the circle fit
    problem. See [1]_ for the derivation.

    References
    ----------
    .. [1]  Maia, N.M.M., Silva, J.M.M. et al, Theoretical and Experimental
       Modal Analysis, p221, Research Studies Press, 1997.
    """
    # Use the method from "Theoretical and Experimental Modal Analysis" p221
    # Set up the matrices
    xs = np.sum(x)
    ys = np.sum(y)
    xx = np.square(x).sum()
    yy = np.square(y).sum()
    xy = np.sum(x*y)
    L = x.size
    xxx = np.sum(x*np.square(x))
    yyy = np.sum(y*np.square(y))
    xyy = np.sum(x*np.square(y))
    yxx = np.sum(y*np.square(x))

    A = np.asarray([[xx, xy, -xs],
                    [xy, yy, -ys],
                    [-xs, -ys, L]])

    B = np.asarray([[-(xxx + xyy)],
                    [-(yyy + yxx)],
                    [xx + yy]])

    # Solve the equation
    v = np.linalg.solve(A, B)

    # Find the circle parameters
    x0 = v[0]/-2
    y0 = v[1]/-2
    R0 = np.sqrt(v[2] + x0**2 + y0**2)
    return x0, y0, R0


class CircleFitWidget(QWidget):
    def __init__(self, parent=None):
        super().__init__()

        self.w = np.zeros(1)
        self.tf = np.zeros(1)
        self.channels = []
        self.current_peak = 0

        self.init_ui()

    def init_ui(self):
        # # Transfer function plot
        self.transfer_function_plotwidget = \
            InteractivePlotWidget(parent=self,
                                  title="Transfer Function",
                                  labels={'bottom': ("Frequency", "Hz"),
                                          'left': ("Transfer Function", "dB")})

        self.transfer_function_plot = \
            self.transfer_function_plotwidget.getPlotItem()

        self.transfer_function_plotwidget.sig_region_changed.connect(self.update_from_region)

        # # Nyquist plot
        self.nyquist_plotwidget = \
            InteractivePlotWidget(parent=self,
                                  show_region=False,
                                  show_crosshair=False,
                                  title="Circle fit",
                                  labels={'bottom': ("Re"),
                                          'left': ("Im")})

        self.nyquist_plot = self.nyquist_plotwidget.getPlotItem()
        self.nyquist_plot.setLimits(xMin=None, xMax=None,
                                    yMin=None, yMax=None,
                                    minXRange=None, maxXRange=None,
                                    minYRange=None, maxYRange=None)
        self.nyquist_plot.setAspectLocked(lock=True, ratio=1)
        self.nyquist_plot.showGrid(x=True, y=True)

        # # Create the items for the plots
        # The item for the raw transfer function
        self.transfer_function_list = []

        # The item for the reconstructed transfer function
        self.constructed_transfer_function = \
            pg.PlotDataItem(pen=pg.mkPen(width=3, style=Qt.DashLine))
        self.transfer_function_plot.addItem(self.constructed_transfer_function)

        # The item for the fitted peaks
        self.peaks = []

        # The item for the raw circle plot
        self.nyquist_plot_list = []

        # The item for the fitted circles
        self.nyquist_plot_peaks_list = []

        # # Table of results
        self.results = CircleFitResults(self)
        self.results.sig_selected_peak_changed.connect(self.set_current_peak)
        self.results.sig_recalculate_fit.connect(self.update_from_region)
        self.results.sig_add_new_peak.connect(self.add_new_peak)
        self.results.sig_manual_parameter_change.connect(self.update_from_table)

        # # Widget layout
        layout = QGridLayout()
        layout.addWidget(self.transfer_function_plotwidget, 0, 0, 1, 2)
        layout.addWidget(self.results, 1, 0)
        layout.addWidget(self.nyquist_plotwidget, 1, 1)
        self.setLayout(layout)

        self.setWindowTitle('Circle fit')
        self.show()

    def test_slot(*args, **kwargs):
        print("Test slot: {}, {}".format(args, kwargs))

    def show_transfer_fn(self, visible=True):
        print("Setting transfer function visible to " + str(visible))
        self.constructed_transfer_fn.setVisible(visible)

    def construct_transfer_fn(self):
        pass

    def add_new_peak(self):
        lower, upper = self.transfer_function_plotwidget.getRegionBounds()
        self.update_from_region(lower, upper)

    def update_from_region(self, region_lower_bound, region_upper_bound):
        for i, channel in enumerate(self.channels):
            self.freq = channel.data("frequency")
            self.w = channel.data("omega")
            self.tf = channel.data("TF")
            self.transfer_function_type = channel.transfer_function_type

            f_in_region = (self.freq >= region_lower_bound) \
                              & (self.freq <= region_upper_bound)

            self.w_reg = np.extract(f_in_region, self.w)
            self.tf_reg = np.extract(f_in_region, self.tf)

            try:
                # Recalculate the geometric circle fit
                self.x0, self.y0, self.R0 = fit_circle_to_data(self.tf_reg.real,
                                                               self.tf_reg.imag)
            except:
                print("Error in fitting geometric circle.")
                traceback.print_exc()
            # Recalculate the parameters
            try:
                wr, zr, cr, phi = self.sdof_get_parameters()

                # Update the results table
                self.results.set_omega(self.current_peak, i, wr)
                self.results.set_damping(self.current_peak, i, zr)
                self.results.set_amplitude(self.current_peak, i, cr)
                self.results.set_phase_rad(self.current_peak, i, phi)

            except:
                print("Error in calculating parameters.")
                traceback.print_exc()

            wr = self.results.get_omega(self.current_peak, i)
            zr = self.results.get_damping(self.current_peak, i)
            cr = self.results.get_amplitude(self.current_peak, i)
            phi = self.results.get_phase_rad(self.current_peak, i)

            # Update the peak
            peak = sdof_modal_peak(self.w, wr, zr, cr, phi)
            self.peaks[i].setData(self.freq, to_dB(np.abs(peak)))
            self.nyquist_plot_peaks_list[i].setData(peak.real, peak.imag)

            # Update what is displayed on the nyquist plot
            self.nyquist_plot_list[i].setData(self.tf_reg.real, self.tf_reg.imag)

            self.nyquist_plot.autoRange()

    def update_from_table(self):
        for i, channel in enumerate(self.channels):
            wr = self.results.get_omega(self.current_peak, i)
            zr = self.results.get_damping(self.current_peak, i)
            cr = self.results.get_amplitude(self.current_peak, i)
            phi = self.results.get_phase_rad(self.current_peak, i)

            # Update the peak
            peak = sdof_modal_peak(self.w, wr, zr, cr, phi)
            self.peaks[i].setData(self.freq, to_dB(np.abs(peak)))
            self.nyquist_plot_peaks_list[i].setData(peak.real, peak.imag)

    def refresh_nyquist_plot(self):
        """Clear the nyquist plot and add the items back in."""
        self.nyquist_plotwidget.clear()

        for item in self.nyquist_plot_list:
            self.nyquist_plot.addItem(item)

        for item in self.nyquist_plot_peaks_list:
            self.nyquist_plot.addItem(item)

        self.nyquist_plot.autoRange()

    def refresh_transfer_function_plot(self):
        """Clear the transfer function plot and add the items back in."""
        self.transfer_function_plotwidget.clear()

        self.transfer_function_plot.addItem(self.constructed_transfer_function)

        for item in self.transfer_function_list:
            self.transfer_function_plot.addItem(item)

        for item in self.peaks:
            self.transfer_function_plot.addItem(item)

        self.transfer_function_plot.autoRange()

    def fitted_sdof_peak(self, w, wr, zr, cr, phi):
        """An SDOF modal peak fitted to the data using the geometric circle."""
        if self.transfer_function_type == 'displacement':
            return self.x0 + 1j*self.y0 - self.R0*np.exp(1j*(phi - np.pi/2))\
                + sdof_modal_peak(w, wr, zr, cr, phi)

        if self.transfer_function_type == 'velocity':
            return self.x0 + 1j*self.y0 - self.R0*np.exp(1j*phi)\
                + 1j*w*sdof_modal_peak(w, wr, zr, cr, phi)

        if self.transfer_function_type == 'acceleration':
            return self.x0 + 1j*self.y0 - self.R0*np.exp(1j*(phi + np.pi/2))\
                -w**2*sdof_modal_peak(w, wr, zr, cr, phi)

    def residuals(self, parameters):
        """The error function for least squares fitting."""
        wr = parameters[0]
        zr = parameters[1]
        cr = parameters[2]
        phi = parameters[3]
        if cr < 0:
            cr *= -1
            phi = (phi + np.pi) % np.pi
        f = self.tf_reg - self.fitted_sdof_peak(self.w_reg, wr, zr, cr, phi)
        #return np.append(tf_real, tf_imag) - np.append(f.real, f.imag)
        return np.sum(f*f.conj()).real

    def sdof_get_parameters(self):
        """Fit a SDOF peak to the data with a least squares fit, using values
        from the current peak as a first guess."""
        # # Find initial parameters for curve fitting
        # Find where the peak is - the maximum magnitude of the amplitude
        # within the region
        i = np.where(np.abs(self.tf_reg) == np.abs(self.tf_reg).max())[0][0]
        # Take the frequency at the max amplitude as a
        # first resonant frequency guess
        wr0 = self.w_reg[i]
        # Take the max amplitude as a first guess for the modal constant
        cr0 = np.abs(self.tf_reg[i])
        phi0 = np.angle(self.tf_reg[i])
        # First guess of damping factor of 1% (Q of 100)
        zr0 = 0.01

        # # Least squares fit
        parameters0 = [wr0, zr0, cr0, phi0]
        #wr, zr, cr, phi = leastsq(self.residuals, parameters0,
        #                          args=(self.w_reg, self.tf_reg.real, self.tf_reg.imag))[0]
        wr, zr, cr, phi = minimize(self.residuals, parameters0).x

        return wr, zr, cr, phi

    def set_selected_channels(self, selected_channels):
        """Update which channels are plotted."""
        # If no channel list is given
        if not selected_channels:
            self.channels = []
        else:
            self.channels = selected_channels
            self.results.channels = selected_channels

        # # Populate the plot lists
        self.transfer_function_list = []
        self.nyquist_plot_list = []
        self.peaks = []
        self.nyquist_plot_peaks_list = []
        for channel in self.channels:
            transfer_function = pg.PlotDataItem(channel.data("frequency"),
                                                to_dB(np.abs(channel.data("TF"))),
                                                pen=channel.colour)
            self.transfer_function_list.append(transfer_function)

            nyquist_plot = pg.PlotDataItem(channel.data("TF").real,
                                           channel.data("TF").imag,
                                           pen=None,
                                           symbol='o',
                                           symbolPen=None,
                                           symbolBrush=pg.mkBrush(channel.colour),
                                           symbolSize=7)
            self.nyquist_plot_list.append(nyquist_plot)

            self.peaks.append(pg.PlotDataItem(pen='k'))
            self.nyquist_plot_peaks_list.append(pg.PlotDataItem(pen='k'))

        self.refresh_transfer_function_plot()
        self.refresh_nyquist_plot()

    def set_current_peak(self, current_peak):
        self.current_peak = current_peak

class CircleFitResults(QGroupBox):
    """
    The tree displaying the Circle Fit results and the parameters used for
    automatically fitting the circle.

    Attributes
    ----------
    sig_selected_peak_changed : pyqtSignal(int)
        The signal emitted when the selected peak is changed.
    sig_recalculate_fit : pyqtSignal
        Signal emitted when the fit is forced to be recalculated.
    sig_add_new_peak : pyqtSignal
        Signal emitted when a new peak is added.
    sig_manual_parameter_change : pyqtSignal
        Signal emitted when a parameter is changed by hand.
    """

    sig_selected_peak_changed = pyqtSignal(int)
    sig_recalculate_fit = pyqtSignal()
    sig_add_new_peak = pyqtSignal()
    sig_manual_parameter_change = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__("Results", parent)

        self.channels = []
        self.num_peaks = 0

        self.init_ui()

    def init_ui(self):
        # # Tree for values
        self.tree = QTreeWidget()
        self.tree.setHeaderLabels(["Name",
                                   "Frequency (Hz)",
                                   "Damping ratio",
                                   "Amplitude",
                                   "Phase (deg)",
                                   "Select"])

        self.tree.setStyleSheet("QTreeWidget::item:has-children "
                                "{ background-color : palette(mid);}")

        # # Tree for autofit parameters
        self.autofit_tree = QTreeWidget()
        self.autofit_tree.setHeaderLabels(["Name",
                                           "Frequency (Hz)",
                                           "Damping ratio",
                                           "Amplitude",
                                           "Phase (deg)",
                                           "Select"])

        self.autofit_tree.hide()
        self.autofit_tree.setStyleSheet("QTreeWidget::item:has-children "
                                        "{ background-color : palette(mid);}")

        # Connect the two trees together, so the views look identical
        self.autofit_tree.itemCollapsed.connect(self.on_item_collapsed)
        self.autofit_tree.itemExpanded.connect(self.on_item_expanded)
        self.tree.itemCollapsed.connect(self.on_item_collapsed)
        self.tree.itemExpanded.connect(self.on_item_expanded)

        # # Controls
        self.add_peak_btn = QPushButton("Add new peak", self)
        self.add_peak_btn.clicked.connect(self.add_peak)

        self.delete_selected_btn = QPushButton("Delete selected", self)
        self.delete_selected_btn.clicked.connect(self.delete_selected)

        self.view_autofit_btn = QPushButton("View autofit parameters", self)
        self.view_autofit_btn.clicked.connect(self.toggle_view)

        self.reset_to_autofit_btn = QPushButton("Reset all to auto", self)
        self.reset_to_autofit_btn.clicked.connect(self.reset_to_autofit)
        self.reset_to_autofit_btn.hide()

        controls1 = QGridLayout()
        controls1.addWidget(self.add_peak_btn, 0, 0)
        controls1.setColumnStretch(1, 1)
        controls1.addWidget(self.delete_selected_btn, 0, 2)

        controls2 = QGridLayout()
        controls2.setColumnStretch(0, 1)
        controls2.addWidget(self.reset_to_autofit_btn, 0, 1)
        controls2.addWidget(self.view_autofit_btn, 0, 2)

        # # Layout
        layout = QGridLayout()
        layout.addLayout(controls2, 0, 0)
        layout.addWidget(self.tree, 1, 0)
        layout.addWidget(self.autofit_tree, 1, 0)
        layout.addLayout(controls1, 2, 0)

        self.setLayout(layout)

    def add_peak(self):
        """Add a top-level item for a new peak to the tree, with children for
        each channel."""
        # Create the parent item for the peak
        peak_item = QTreeWidgetItem(self.tree,
                                    ["Peak {}".format(self.num_peaks),
                                     "", "", "", "", ""])
        autofit_peak_item = QTreeWidgetItem(self.autofit_tree,
                                            ["Peak {}".format(self.num_peaks),
                                             "", "", "", "", ""])

        # Put a radio button in column 5 in the tree
        radio_btn = QRadioButton()
        radio_btn.toggled.connect(self.on_selected_peak_changed)
        self.tree.setItemWidget(peak_item, 5, radio_btn)
        radio_btn.setChecked(True)

        for col in [1, 2, 3, 4]:
            # Put comboboxes in the autofit tree
            combobox = QComboBox()
            combobox.addItems(["Auto", "Manual"])
            combobox.currentIndexChanged.connect(self.update_peak_average)
            self.autofit_tree.setItemWidget(autofit_peak_item, col, combobox)

        # Put spinboxes in the tree
        freq_spinbox = QDoubleSpinBox()
        freq_spinbox.setRange(0, 9e99)
        freq_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
        self.tree.setItemWidget(peak_item, 1, freq_spinbox)
        damping_spinbox = QDoubleSpinBox()
        damping_spinbox.setRange(0, 9e99)
        damping_spinbox.setDecimals(5)
        damping_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
        self.tree.setItemWidget(peak_item, 2, damping_spinbox)
        amp_spinbox = QDoubleSpinBox()
        amp_spinbox.setRange(0, 9e99)
        amp_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
        self.tree.setItemWidget(peak_item, 3, amp_spinbox)
        phase_spinbox = QDoubleSpinBox()
        phase_spinbox.setRange(-180, 180)
        phase_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
        phase_spinbox.setWrapping(True)
        self.tree.setItemWidget(peak_item, 4, phase_spinbox)

        # Create the child items for each channel for this peak if there's
        # more than one channel
        if len(self.channels) > 1:
            for i in range(len(self.channels)):
                channel_item = QTreeWidgetItem(peak_item,
                                               ["Channel {}".format(i),
                                                "", "", "", "", ""])
                autofit_channel_item = QTreeWidgetItem(autofit_peak_item,
                                                       ["Channel {}".format(i),
                                                        "", "", "", "", ""])
                for col in [1, 2, 3, 4]:
                    # Put comboboxes in the autofit tree
                    combobox = QComboBox()
                    combobox.addItems(["Auto", "Manual"])
                    combobox.currentIndexChanged.connect(self.update_peak_average)
                    self.autofit_tree.setItemWidget(autofit_channel_item,
                                                    col, combobox)

                freq_spinbox = QDoubleSpinBox()
                freq_spinbox.setRange(0, 9e99)
                freq_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
                freq_spinbox.valueChanged.connect(self.update_peak_average)
                self.tree.setItemWidget(channel_item, 1, freq_spinbox)
                damping_spinbox = QDoubleSpinBox()
                damping_spinbox.setRange(0, 9e99)
                damping_spinbox.setDecimals(5)
                damping_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
                damping_spinbox.valueChanged.connect(self.update_peak_average)
                self.tree.setItemWidget(channel_item, 2, damping_spinbox)
                amp_spinbox = QDoubleSpinBox()
                amp_spinbox.setRange(0, 9e99)
                amp_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
                amp_spinbox.valueChanged.connect(self.update_peak_average)
                self.tree.setItemWidget(channel_item, 3, amp_spinbox)
                phase_spinbox = QDoubleSpinBox()
                phase_spinbox.setRange(-180, 180)
                phase_spinbox.setWrapping(True)
                phase_spinbox.valueChanged.connect(self.sig_manual_parameter_change.emit)
                phase_spinbox.valueChanged.connect(self.update_peak_average)
                self.tree.setItemWidget(channel_item, 4, phase_spinbox)

        # Register that we've added another peak
        self.num_peaks += 1
        self.sig_add_new_peak.emit()

    def delete_selected(self):
        """Delete the item that is currently selected."""
        for i in range(self.tree.topLevelItemCount()):
            # If the radio button is checked
            peak_item = self.tree.topLevelItem(i)
            if peak_item is not None:
                if self.tree.itemWidget(peak_item, 5).isChecked():
                    # Delete this item
                    self.tree.takeTopLevelItem(i)
                    self.autofit_tree.takeTopLevelItem(i)
                    # self.num_peaks -= 1

    def toggle_view(self):
        if self.tree.isVisible():
            self.tree.hide()
            self.autofit_tree.show()
            self.reset_to_autofit_btn.show()
            self.view_autofit_btn.setText("View parameter values")
        else:
            self.tree.show()
            self.autofit_tree.hide()
            self.reset_to_autofit_btn.hide()
            self.view_autofit_btn.setText("View autofit parameters")

    def on_item_collapsed(self, item):
        index = self.sender().indexOfTopLevelItem(item)
        self.tree.collapseItem(self.tree.topLevelItem(index))
        self.autofit_tree.collapseItem(self.autofit_tree.topLevelItem(index))

    def on_item_expanded(self, item):
        index = self.sender().indexOfTopLevelItem(item)
        self.tree.expandItem(self.tree.topLevelItem(index))
        self.autofit_tree.expandItem(self.autofit_tree.topLevelItem(index))

    def on_selected_peak_changed(self, checked):
        for i in range(self.tree.topLevelItemCount()):
            # If the radio button in this row is the sender
            peak_item = self.tree.topLevelItem(i)
            if peak_item is not None:
                if self.tree.itemWidget(peak_item, 5) == self.sender():
                    if checked:
                        print("Selected peak: " + str(i))
                        self.sig_selected_peak_changed.emit(i)

    def reset_to_autofit(self):
        """Reset all parameters to be automatically adjusted."""
        for peak_number in range(self.num_peaks):
            peak_item = self.autofit_tree.topLevelItem(peak_number)

            for col in [1, 2, 3, 4]:
                self.autofit_tree.itemWidget(peak_item, col).setCurrentIndex(0)

            if len(self.channels) > 1:
                for channel_number in range(len(self.channels)):
                    channel_item = peak_item.child(channel_number)
                    for col in [1, 2, 3, 4]:
                        self.autofit_tree.itemWidget(channel_item, col).setCurrentIndex(0)

        self.sig_recalculate_fit.emit()

    def update_peak_average(self):
        """Set the parameter values displayed for the peak to the average of
        all the channel values for each parameter."""
        for peak_number in range(self.num_peaks):
            # Get the peak item
            peak_item = self.tree.topLevelItem(peak_number)
            autofit_item = self.autofit_tree.topLevelItem(peak_number)

            if peak_item is not None:
                # Find the average values of all the channels
                avg_freq = 0
                avg_damping = 0
                avg_amplitude = 0
                avg_phase_deg = 0

                for channel_number in range(len(self.channels)):
                    avg_freq += self.get_frequency(peak_number, channel_number)
                    avg_damping += self.get_damping(peak_number, channel_number)
                    avg_amplitude += self.get_amplitude(peak_number, channel_number)
                    avg_phase_deg += self.get_phase_deg(peak_number, channel_number)

                avg_freq /= len(self.channels)
                avg_damping /= len(self.channels)
                avg_amplitude /= len(self.channels)
                avg_phase_deg /= len(self.channels)

                # Set the peak item to display the averages
                if self.autofit_tree.itemWidget(autofit_item, 1).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 1).setValue(avg_freq)
                if self.autofit_tree.itemWidget(autofit_item, 2).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 2).setValue(avg_damping)
                if self.autofit_tree.itemWidget(autofit_item, 3).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 3).setValue(avg_amplitude)
                if self.autofit_tree.itemWidget(autofit_item, 4).currentText() == "Auto":
                    self.tree.itemWidget(peak_item, 4).setValue(avg_phase_deg)

    def get_frequency(self, peak_number, channel_number=None):
        """Return the resonant frequency (Hz) of the peak given by
        *peak_number*. If *channel_number* is given, return the resonant
        frequency of the given peak in the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 1)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 1)
                return spinbox.value()
        else:
            return 0

    def get_omega(self, peak_number, channel_number=None):
        """Return the resonant frequency (rads) of the peak given by
        *peak_number*. If *channel_number* is given, return the resonant
        frequency of the given peak in the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 1)
                return spinbox.value() * 2*np.pi
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 1)
                return spinbox.value() * 2*np.pi
        else:
            return 0

    def get_damping(self, peak_number, channel_number=None):
        """Return the damping ratio of the peak given by *peak_number*. If
        *channel_number* is given, return the damping ratio of the given peak
        in the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 2)
                #return 1/(2*spinbox.value())
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 2)
                #return 1/(2*spinbox.value())
                return spinbox.value()
        else:
            return 0

    """
    def get_q(self, peak_number, channel_number=None):
    """        """Return the Q factor of the peak given by *peak_number*. If
        *channel_number* is given, return the damping ratio of the given peak
        in the given channel.""""""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 2)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 2)
                return spinbox.value()
        else:
            return 0


    def get_amplitude(self, peak_number, channel_number=None):
    """    """Return the amplitude of the peak given by *peak_number*. If
        *channel_number* is given, return the amplitude of the given peak in
        the given channel.""""""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 3)
                return from_dB(spinbox.value())
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 3)
                return from_dB(spinbox.value())
        else:
            return 0
    """
    def get_amplitude(self, peak_number, channel_number=None):
        """Return the amplitude of the peak given by *peak_number*. If
        *channel_number* is given, return the amplitude of the given peak in
        the given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 3)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 3)
                return spinbox.value()
        else:
            return 0

    def get_phase_deg(self, peak_number, channel_number=None):
        """Return the phase (deg) of the peak given by *peak_number*. If
        *channel_number* is given, return the phase of the given peak in the
        given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 4)
                return spinbox.value()
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 4)
                return spinbox.value()
        else:
            return 0

    def get_phase_rad(self, peak_number, channel_number=None):
        """Return the phase (rad) of the peak given by *peak_number*. If
        *channel_number* is given, return the phase of the given peak in the
        given channel."""
        peak_item = self.tree.topLevelItem(peak_number)
        if peak_item is not None:
            if channel_number is None:
                spinbox = self.tree.itemWidget(peak_item, 4)
                return np.deg2rad(spinbox.value())
            else:
                channel_item = peak_item.child(channel_number)
                spinbox = self.tree.itemWidget(channel_item, 4)
                return np.deg2rad(spinbox.value())
        else:
            return 0

    def set_frequency(self, peak_number, channel_number=None, value=0):
        """Set the frequency (Hz) of a given peak to *value*. If *channel_number* is
        given, set the frequency of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 1)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 1)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 1)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 1)
                    spinbox.setValue(value)

    def set_omega(self, peak_number, channel_number=None, value=0):
        """Set the frequency (rad) of a given peak to *value*. If *channel_number* is
        given, set the frequency of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 1)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 1)
                    spinbox.setValue(value / (2*np.pi))
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 1)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 1)
                    spinbox.setValue(value / (2*np.pi))

    def set_damping(self, peak_number, channel_number=None, value=0):
        """Set the damping ratio of a given peak to *value*. If *channel_number* is
        given, set the damping ratio of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 2)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 2)
                    spinbox.setValue(1/(2*value))
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 2)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 2)
                    spinbox.setValue(1/(2*value))

    def set_q(self, peak_number, channel_number=None, value=0):
        """Set the q factor of a given peak to *value*. If *channel_number* is
        given, set the damping ratio of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 2)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 2)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 2)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 2)
                    spinbox.setValue(value)

    def set_amplitude(self, peak_number, channel_number=None, value=0):
        """Set the amplitude of a given peak to *value*. If *channel_number* is
        given, set the amplitude of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 3)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 3)
                    spinbox.setValue(to_dB(value))
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 3)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 3)
                    spinbox.setValue(to_dB(value))

    def set_amplitude_dB(self, peak_number, channel_number=None, value=0):
        """Set the amplitude of a given peak to *value*. If *channel_number* is
        given, set the amplitude of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 3)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 3)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 3)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 3)
                    spinbox.setValue(value)

    def set_phase_deg(self, peak_number, channel_number=None, value=0):
        """Set the phase of a given peak to *value*. If *channel_number* is
        given, set the phase of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 4)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 4)
                    spinbox.setValue(value)
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 4)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 4)
                    spinbox.setValue(value)

    def set_phase_rad(self, peak_number, channel_number=None, value=0):
        """Set the phase of a given peak to *value*. If *channel_number* is
        given, set the phase of the given peak in the given channel."""
        # Get the top level items
        peak_item = self.tree.topLevelItem(peak_number)
        autofit_peak_item = self.autofit_tree.topLevelItem(peak_number)
        # Check they exist
        if peak_item is not None:
            # If they have children
            if channel_number is not None:
                # Get the channel items
                channel_item = peak_item.child(channel_number)
                autofit_channel_item = autofit_peak_item.child(channel_number)
                # Set the value only if the combobox is set to autofit
                combobox = \
                    self.autofit_tree.itemWidget(autofit_channel_item, 4)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(channel_item, 4)
                    spinbox.setValue(np.rad2deg(value))
            else:
                # Set the value only if the combobox is set to autofit
                combobox = self.autofit_tree.itemWidget(autofit_peak_item, 4)
                if combobox.currentText() == "Auto":
                    spinbox = self.tree.itemWidget(peak_item, 4)
                    spinbox.setValue(np.rad2deg(value))


class CircleFitToolbox(Toolbox):
    """
    The Toolbox for the CircleFitWidget.

    This Toolbox contains the tools for controlling the circle fit. It has
    two tabs: 'Transfer Function', for tools relating to the construction
    of a transfer function, and 'Autofit Controls', which contains tools
    for controlling how the circle is fit to the data.

    Attributes
    ----------
    sig_construct_transfer_fn : pyqtSignal
      The signal emitted when a new transfer function is to be constructed.
    sig_show_transfer_fn : pyqtSignal(bool)
      The signal emitted when the visibility of the transfer function is
      changed. Format (visible).
    """

    sig_construct_transfer_fn = pyqtSignal()
    sig_show_transfer_fn = pyqtSignal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.init_ui()

    def init_ui(self):
        self.init_transfer_function_tab()

    def init_transfer_function_tab(self):
        self.transfer_function_tab = QWidget()
        transfer_function_tab_layout = QGridLayout()

        self.construct_transfer_fn_btn = QPushButton()
        self.construct_transfer_fn_btn.setText("Construct transfer function")
        self.construct_transfer_fn_btn.clicked.connect(self.sig_construct_transfer_fn.emit)
        transfer_function_tab_layout.addWidget(self.construct_transfer_fn_btn, 0, 0)

        self.show_transfer_fn_checkbox = QCheckBox()
        self.show_transfer_fn_checkbox.setText("Show transfer function")
        self.show_transfer_fn_checkbox.stateChanged.connect(self.sig_show_transfer_fn.emit)
        transfer_function_tab_layout.addWidget(self.show_transfer_fn_checkbox, 1, 0)

        self.transfer_function_tab.setLayout(transfer_function_tab_layout)
        transfer_function_tab_layout.setColumnStretch(1, 1)
        transfer_function_tab_layout.setRowStretch(2, 1)

        self.addTab(self.transfer_function_tab, "Transfer function")


if __name__ == '__main__':
    CurrentWorkspace = Workspace()
    app = 0

    app = QApplication(sys.argv)
    c = CircleFitWidget()
    c.CurrentWorkspace = CurrentWorkspace
    c.showMaximized()

    cs = ChannelSet()

    #c.load_tf(cs)
    import_from_mat("../../tests/transfer_function_grid.mat", cs)
    c.transfer_function_type = 'acceleration'
    c.set_selected_channels(cs.channels)
    #c.set_data(np.linspace(0, cs.channel_metadata(0, "sample_rate"), a.size), a)
    #"""
    sys.exit(app.exec_())



import weakref
import sys
if __name__ == '__main__':
    sys.path.append('../../')
import numpy as np
import pyqtgraph as pg
from pyqtgraph import ImageItem
from cued_datalogger.api.pyqt_widgets import matplotlib_lookup_table
from PyQt5.QtWidgets import(QWidget,QMenu,QAction,QActionGroup,QWidgetAction,QGridLayout,
                            QCheckBox,QRadioButton,QLineEdit,QSpinBox,QComboBox,
                            QLabel, QApplication, QVBoxLayout, QHBoxLayout, QPushButton)
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtCore import QMetaObject,QSize,QCoreApplication, QTimer, pyqtSignal


class InteractivePlotWidget(QWidget):
    """A QWidget containing a :class:`CustomPlotWidget` with mouse tracking
    crosshairs, a :class:`LinearRegionItem`, and spinboxes
    to display and control the values of the bounds of the linear region.
    Any additional arguments to :method:`__init__` are passed to the
    CustomPlotWidget.

    Attributes
    ----------
    PlotWidget : pg.PlotWidget
        The PlotWidget contained in the InteractivePlotWidget.
    ViewBox : pg.ViewBox
        The ViewBox contained in the InteractivePlotWidget.
    region : pg.LinearRegionItem
        The LinearRegionItem contained in the InteractivePlotWidget.
    vline : pg.InfiniteLine
        Vertical mouse-tracking line.
    hline : pg.InfiniteLine
        Horizontal mouse-tracking line.
    label : pg.LabelItem
        LabelItem displaying current mouse position.
    lower_box : QSpinBox
        QSpinBox displaying lower bound of :attr:`region`.
    upper_box : QSpinBox
        QSpinBox displaying upper bound of :attr:`region`.
    zoom_btn : QPushButton
        Press to zoom to the :attr:`region` with a set amount of padding.
    show_region : bool
        Controls whether the region is displayed.
    show_crosshair : bool
        Controls whether the crosshair is displayed.
    sig_region_changed : pyqtSignal([int, int])
        The signal emitted when the region is changed.
    """

    sig_region_changed = pyqtSignal([float, float])

    def __init__(self, parent=None,
                 show_region=True, show_crosshair=True, show_label=True,
                 *args, **kwargs):
        self.parent = parent
        self.show_region = show_region
        self.show_crosshair = show_crosshair
        self.show_label = show_label

        super().__init__(parent)

        layout = QVBoxLayout(self)

        # # Set up the PlotWidget
        self.PlotWidget = CustomPlotWidget(self, *args, **kwargs)

        self.PlotItem = self.PlotWidget.getPlotItem()
        self.PlotItem.disableAutoRange()

        self.ViewBox = self.PlotWidget.getViewBox()
        self.ViewBox.menu.show_crosshair_action.setChecked(self.show_crosshair)
        self.ViewBox.menu.show_region_action.setChecked(self.show_region)
        self.ViewBox.menu.show_label_action.setChecked(self.show_label)


        self.vline = pg.InfiniteLine(angle=90)
        self.hline = pg.InfiniteLine(angle=0)

        self.region = pg.LinearRegionItem(bounds=[0, None])
        self.region.sigRegionChanged.connect(self.updateBoxFromRegion)

        layout.addWidget(self.PlotWidget)

        self.label = pg.LabelItem(angle = 0)
        self.label.setParentItem(self.ViewBox)

        self.proxy = pg.SignalProxy(self.PlotWidget.scene().sigMouseMoved, rateLimit=30, slot=self.mouseMoved)

        self.ViewBox.menu.sig_show_region.connect(self.set_show_region)
        self.ViewBox.menu.sig_show_crosshair.connect(self.set_show_crosshair)
        self.ViewBox.menu.sig_show_label.connect(self.set_show_label)

        # # Set up the region controls
        control_layout = QHBoxLayout()
        self.lower_box = pg.SpinBox(self, bounds=(None, None))
        self.lower_box_proxy = pg.SignalProxy(self.lower_box.valueChanged, rateLimit=30, slot=self.on_region_changed)
        self.upper_box = pg.SpinBox(self, bounds=(None, None))
        self.upper_box_proxy = pg.SignalProxy(self.upper_box.valueChanged, rateLimit=30, slot=self.on_region_changed)
        self.zoom_btn = QPushButton('Zoom', self)
        self.zoom_btn.clicked.connect(self.zoomToRegion)

        control_layout.addWidget(QLabel('Lower', self))
        control_layout.addWidget(self.lower_box)
        control_layout.addWidget(QLabel('Upper', self))
        control_layout.addWidget(self.upper_box)
        control_layout.addWidget(self.zoom_btn)
        layout.addLayout(control_layout)

        # # Create a QTimer for smooth updating of the region
        self.updatetimer = QTimer(self)
        self.updatetimer.timeout.connect(self.updateRegionFromBox)
        self.updatetimer.start(20)

        self.clear()

    def on_region_changed(self):
        lower = self.lower_box.value()
        upper = self.upper_box.value()
        self.sig_region_changed.emit(lower, upper)

    def mouseMoved(self, mouse_moved_event):
        """
        Update the mouse position with crosshair
        """
        mouse_position = mouse_moved_event[0]  ## using signal proxy turns original arguments into a tuple
        # If the mouse is in the PlotItem
        if self.PlotItem.sceneBoundingRect().contains(mouse_position):
            # Convert it to the coordinate system
            try:
                mousePoint = self.ViewBox.mapSceneToView(mouse_position)
                # Update the label
                self.label.setText((("<span style='font-size: 12pt;color: black'>"
                                     "x=%0.4f,"
                                     "<span style='color: red'>y1=%0.4f</span>" )
                                    % (mousePoint.x(), mousePoint.y()) ))
                self.vline.setPos(mousePoint.x())
                self.hline.setPos(mousePoint.y())
            except np.linalg.LinAlgError:
                print("Warning: Exception raised in calculating mouse position"
                      " (np.linalg.LinAlgError)")

    def clear(self):
        """Clear the PlotItem and add the default items back in."""
        self.PlotItem.clear()
        if self.show_crosshair:
            self.PlotItem.addItem(self.vline)
            self.PlotItem.addItem(self.hline)
        if self.show_region:
            self.PlotItem.addItem(self.region)

    def updateRegionFromBox(self):
        # Get the bounds of the region as defined by the spinboxes
        region_bounds = [self.lower_box.value(),
                         self.upper_box.value()]
        # Sort them so that the lower one is first
        region_bounds.sort()
        # Update the region
        self.region.setRegion(region_bounds)
        # Update the spinboxes
        self.lower_box.setValue(region_bounds[0])
        self.upper_box.setValue(region_bounds[1])

    def updateBoxFromRegion(self):
        # Get the region bounds as defined by the region
        region_bounds = list(self.region.getRegion())
        # Sort it so the lowest is first
        region_bounds.sort()
        # Set the spinboxes to reflect the bounds of the region
        self.lower_box.setValue(region_bounds[0])
        self.upper_box.setValue(region_bounds[1])

    def zoomToRegion(self, padding=0.1):
        """Zoom to the region, with given padding."""
        pos = self.region.getRegion()
        self.PlotItem.setXRange(pos[0],pos[1],padding=padding)

    def getRegionBounds(self):
        """Return the lower and upper bounds of the region."""
        return self.lower_box.value(), self.upper_box.value()

    def closeEvent(self, close_event):
        # Tidy everything up when told to close
        #self.proxy.disconnect()
        if self.updatetimer.isActive():
            self.updatetimer.stop()
        close_event.accept()

    def plot(self, x=None, y=None, *args, **kwargs):
        """:func:`update_limits` from the x and y values, then plot
        the data on the plotWidget."""
        self.update_limits(x, y)
        self.PlotWidget.plot(x, y, *args, **kwargs)
        self.ViewBox.autoRange()

    def update_limits(self, x, y):
        """Set the increment of the spinboxes, the limits of zooming and
        scrolling the PlotItem, and move the region to x=0"""

        if x is not None and y is not None:
            # Update the increment of the spinboxes
            self.lower_box.setSingleStep(x.max()/100)
            self.upper_box.setSingleStep(x.max()/100)

            # Set the linear region to be in view
            #self.lower_box.setValue(x.max()*0.4)
            #self.upper_box.setValue(x.max()*0.6)
            #self.lower_box.setValue(0)
            #self.upper_box.setValue(0)

            # Set the limits of the PlotItem
            self.PlotItem.setLimits(xMin=x.min(), xMax=x.max())
            self.PlotItem.setRange(xRange=(x.min(), x.max()),
                                   yRange=(y.min(), y.max()),
                                   padding=0.2)


    def getPlotItem(self):
        """Return the PlotItem (reimplemented from
        :method:`pg.PlotWidget.getPlotItem`)."""
        return self.PlotWidget.getPlotItem()

    def set_show_crosshair(self, show_crosshair):
        """Set whether the crosshair is visible."""
        self.show_crosshair = show_crosshair

        if self.show_crosshair:
            self.PlotItem.addItem(self.vline)
            self.PlotItem.addItem(self.hline)
        else:
            self.PlotItem.removeItem(self.vline)
            self.PlotItem.removeItem(self.hline)

    def set_show_region(self, show_region):
        """Set whether the region is visible."""
        self.show_region = show_region

        if self.show_region:
            self.PlotItem.addItem(self.region)
        else:
            self.PlotItem.removeItem(self.region)

    def set_show_label(self, show_label):
        """Set whether the label is visible."""
        self.show_label = show_label

        if self.show_label:
            self.label.show()
        else:
            self.label.hide()


class CustomPlotWidget(pg.PlotWidget):
    '''
    Reimplementation of the original PlotWidget.
    This is where you would implement options to modify the plots.
    For that, reimplement the ctrlMenu variable. Currently, only certain options
    are removed, rather than adding any new custom actions.
    Also, bring out the PlotItem and ViewBox for easy access.

    Attributes
    ----------
    PlotItem: PlotItem

    ViewBox: ViewBox
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, viewBox=CustomViewBox(cparent=self), **kwargs)

        self.PlotItem = self.getPlotItem()
        self.ViewBox = self.PlotItem.getViewBox()

        # Removing some plot options, HERE!!!!!
        ext_menu = self.PlotItem.ctrlMenu
        ext_submenus = self.PlotItem.subMenus
        ext_menu.removeAction(ext_submenus[1].menuAction())
        ext_menu.removeAction(ext_submenus[2].menuAction())
        ext_menu.removeAction(ext_submenus[3].menuAction())
        ext_menu.removeAction(ext_submenus[5].menuAction())

    def getViewBox(self):
        return self.ViewBox

    def getViewedDataRegion(self, data_for_slice, axis='bottom'):
        """Return the indices of data_for_slice that are currently in the
        visible plot window"""

        # Get the axis limits
        axis_lower, axis_upper = self.PlotItem.getAxis(axis).range

        # The data that is in view is where the data is within the axis limits
        data_in_display = (data_for_slice >= axis_lower) & (data_for_slice <= axis_upper)

        indices_of_data_in_display = np.where(data_in_display)[0]
        return indices_of_data_in_display[0], indices_of_data_in_display[-1]


class CustomViewBox(pg.ViewBox):
    '''
    A custom ViewBox, reimplemented to allow custom right click context menu
    '''
    def __init__(self, cparent=None, *arg, **kwarg):
        super().__init__(*arg,**kwarg)
        self.cparent = cparent
        self.menu = CustomViewMenu(self)

    def raiseContextMenu(self, ev):
        menu = self.getMenu(ev)
        menu.addMenu(self.cparent.getPlotItem().ctrlMenu)
        menu.popup(ev.screenPos().toPoint())

    # Temporarily disable the autorange
    #def autoRange(self, padding= None, items=None):
    #    super().autoRange(padding=padding, items= None)
        #r = self.viewRect()
        #self.setLimits(xMin = r.left(), xMax = r.right())


class CustomViewMenu(QMenu):
    '''
    A custom right click context menu. This is a complete implementation of the
    original context menu. The purpose of these is to allow custom implementation
    of actions. In this case, there is extra options to toggle on/off the crosshair.
    Note that this context menu only affects the view, not the data, i.e. 'Plot Options'.
    For that, you would need to reimplement ctrlMenu in the PlotItem, demonstrated in
    InteractivePlotWidget.
    '''
    sig_show_crosshair = pyqtSignal(bool)
    sig_show_region = pyqtSignal(bool)
    sig_show_label = pyqtSignal(bool)

    def __init__(self, view):
        QMenu.__init__(self)

        self.view = weakref.ref(view)  ## keep weakref to view to avoid circular reference (don't know why, but this prevents the ViewBox from being collected)
        self.valid = False  ## tells us whether the ui needs to be updated
        self.viewMap = weakref.WeakValueDictionary()  ## weakrefs to all views listed in the link combos

        # Autorange option
        self.setTitle("ViewBox options")
        self.autorange_action = QAction("Autorange", self)
        self.autorange_action.triggered.connect(self.autoRange)
        self.addAction(self.autorange_action)

        # Display menu, our custom actions
        self.display_menu = QMenu("Display options")

        self.show_crosshair_action = QAction("Show crosshair", self.display_menu)
        self.show_crosshair_action.setCheckable(True)
        #show_crosshair_action.setChecked(self.view().cparent.parent().show_crosshair)
        self.show_crosshair_action.triggered.connect(self.sig_show_crosshair.emit)
        self.display_menu.addAction(self.show_crosshair_action)

        self.show_region_action = QAction("Show region", self.display_menu)
        self.show_region_action.setCheckable(True)
        #show_region_action.setChecked(self.view().cparent.parent().show_region)
        self.show_region_action.triggered.connect(self.sig_show_region.emit)
        self.display_menu.addAction(self.show_region_action)

        self.show_label_action = QAction("Show Label", self.display_menu)
        self.show_label_action.setCheckable(True)
        #show_label_action.setChecked(self.view().cparent.parent().show_label)
        self.show_label_action.triggered.connect(self.sig_show_label.emit)
        self.display_menu.addAction(self.show_label_action)

        self.addMenu(self.display_menu)

        # One-to-one reimplementation of the original options
        self.axes = []
        self.ctrl = []
        self.widgetGroups = []
        self.dv = QDoubleValidator(self)
        for axis in 'XY':
            m = QMenu()
            m.setTitle("%s Axis" % axis)
            w = QWidget()
            ui = CustomUITemplate()
            ui.setupUi(w)
            a = QWidgetAction(self)
            a.setDefaultWidget(w)
            m.addAction(a)
            self.addMenu(m)
            self.axes.append(m)
            self.ctrl.append(ui)
            self.widgetGroups.append(w)

            connects = [
                (ui.mouseCheck.toggled, 'MouseToggled'),
                (ui.manualRadio.clicked, 'ManualClicked'),
                (ui.minText.editingFinished, 'MinTextChanged'),
                (ui.maxText.editingFinished, 'MaxTextChanged'),
                (ui.autoRadio.clicked, 'AutoClicked'),
                (ui.autoPercentSpin.valueChanged, 'AutoSpinChanged'),
                (ui.linkCombo.currentIndexChanged, 'LinkComboChanged'),
                (ui.autoPanCheck.toggled, 'AutoPanToggled'),
                (ui.visibleOnlyCheck.toggled, 'VisibleOnlyToggled')
            ]

            for sig, fn in connects:
                sig.connect(getattr(self, axis.lower()+fn))

        self.ctrl[0].invertCheck.toggled.connect(self.xInvertToggled)
        self.ctrl[1].invertCheck.toggled.connect(self.yInvertToggled)

        self.leftMenu = QMenu("Mouse Mode")
        group = QActionGroup(self)

        pan = QAction("3 button", self.leftMenu)
        zoom = QAction("1 button", self.leftMenu)
        self.leftMenu.addAction(pan)
        self.leftMenu.addAction(zoom)
        pan.triggered.connect(self.set3ButtonMode)
        zoom.triggered.connect(self.set1ButtonMode)

        pan.setCheckable(True)
        zoom.setCheckable(True)
        pan.setActionGroup(group)
        zoom.setActionGroup(group)
        self.mouseModes = [pan, zoom]
        self.addMenu(self.leftMenu)

        self.view().sigStateChanged.connect(self.viewStateChanged)

        self.updateState()

    def viewStateChanged(self):
        self.valid = False
        if self.ctrl[0].minText.isVisible() or self.ctrl[1].minText.isVisible():
            self.updateState()

    def updateState(self):
        ## Something about the ViewBox has changed; update the menu GUI
        state = self.view().getState(copy=False)
        if state['mouseMode'] == pg.ViewBox.PanMode:
            self.mouseModes[0].setChecked(True)
        else:
            self.mouseModes[1].setChecked(True)

        for i in [0,1]:  # x, y
            tr = state['targetRange'][i]
            self.ctrl[i].minText.setText("%0.5g" % tr[0])
            self.ctrl[i].maxText.setText("%0.5g" % tr[1])
            if state['autoRange'][i] is not False:
                self.ctrl[i].autoRadio.setChecked(True)
                if state['autoRange'][i] is not True:
                    self.ctrl[i].autoPercentSpin.setValue(state['autoRange'][i]*100)
            else:
                self.ctrl[i].manualRadio.setChecked(True)
            self.ctrl[i].mouseCheck.setChecked(state['mouseEnabled'][i])

            ## Update combo to show currently linked view
            c = self.ctrl[i].linkCombo
            c.blockSignals(True)
            try:
                view = state['linkedViews'][i]  ## will always be string or None
                if view is None:
                    view = ''

                ind = c.findText(view)

                if ind == -1:
                    ind = 0
                c.setCurrentIndex(ind)
            finally:
                c.blockSignals(False)

            self.ctrl[i].autoPanCheck.setChecked(state['autoPan'][i])
            self.ctrl[i].visibleOnlyCheck.setChecked(state['autoVisibleOnly'][i])
            xy = ['x', 'y'][i]
            self.ctrl[i].invertCheck.setChecked(state.get(xy+'Inverted', False))

        self.valid = True

    def popup(self, *args):
        if not self.valid:
            self.updateState()
        QMenu.popup(self, *args)

    def autoRange(self):
        self.view().autoRange()  ## don't let signal call this directly--it'll add an unwanted argument

    def xMouseToggled(self, b):
        self.view().setMouseEnabled(x=b)

    def xManualClicked(self):
        self.view().enableAutoRange(pg.ViewBox.XAxis, False)

    def xMinTextChanged(self):
        self.ctrl[0].manualRadio.setChecked(True)
        self.view().setXRange(float(self.ctrl[0].minText.text()), float(self.ctrl[0].maxText.text()), padding=0)

    def xMaxTextChanged(self):
        self.ctrl[0].manualRadio.setChecked(True)
        self.view().setXRange(float(self.ctrl[0].minText.text()), float(self.ctrl[0].maxText.text()), padding=0)

    def xAutoClicked(self):
        val = self.ctrl[0].autoPercentSpin.value() * 0.01
        self.view().enableAutoRange(pg.ViewBox.XAxis, val)

    def xAutoSpinChanged(self, val):
        self.ctrl[0].autoRadio.setChecked(True)
        self.view().enableAutoRange(pg.ViewBox.XAxis, val*0.01)

    def xLinkComboChanged(self, ind):
        self.view().setXLink(str(self.ctrl[0].linkCombo.currentText()))

    def xAutoPanToggled(self, b):
        self.view().setAutoPan(x=b)

    def xVisibleOnlyToggled(self, b):
        self.view().setAutoVisible(x=b)


    def yMouseToggled(self, b):
        self.view().setMouseEnabled(y=b)

    def yManualClicked(self):
        self.view().enableAutoRange(pg.ViewBox.YAxis, False)

    def yMinTextChanged(self):
        self.ctrl[1].manualRadio.setChecked(True)
        self.view().setYRange(float(self.ctrl[1].minText.text()), float(self.ctrl[1].maxText.text()), padding=0)

    def yMaxTextChanged(self):
        self.ctrl[1].manualRadio.setChecked(True)
        self.view().setYRange(float(self.ctrl[1].minText.text()), float(self.ctrl[1].maxText.text()), padding=0)

    def yAutoClicked(self):
        val = self.ctrl[1].autoPercentSpin.value() * 0.01
        self.view().enableAutoRange(pg.ViewBox.YAxis, val)

    def yAutoSpinChanged(self, val):
        self.ctrl[1].autoRadio.setChecked(True)
        self.view().enableAutoRange(pg.ViewBox.YAxis, val*0.01)

    def yLinkComboChanged(self, ind):
        self.view().setYLink(str(self.ctrl[1].linkCombo.currentText()))

    def yAutoPanToggled(self, b):
        self.view().setAutoPan(y=b)

    def yVisibleOnlyToggled(self, b):
        self.view().setAutoVisible(y=b)

    def yInvertToggled(self, b):
        self.view().invertY(b)

    def xInvertToggled(self, b):
        self.view().invertX(b)

    def exportMethod(self):
        act = self.sender()
        self.exportMethods[str(act.text())]()

    def set3ButtonMode(self):
        self.view().setLeftButtonAction('pan')

    def set1ButtonMode(self):
        self.view().setLeftButtonAction('rect')

    def setViewList(self, views):
        names = ['']
        self.viewMap.clear()

        ## generate list of views to show in the link combo
        for v in views:
            name = v.name
            if name is None:  ## unnamed views do not show up in the view list (although they are linkable)
                continue
            names.append(name)
            self.viewMap[name] = v

        for i in [0,1]:
            c = self.ctrl[i].linkCombo
            current = c.currentText()
            c.blockSignals(True)
            changed = True
            try:
                c.clear()
                for name in names:
                    c.addItem(name)
                    if name == current:
                        changed = False
                        c.setCurrentIndex(c.count()-1)
            finally:
                c.blockSignals(False)

            if changed:
                c.setCurrentIndex(0)
                c.currentIndexChanged.emit(c.currentIndex())

class CustomUITemplate(object):
    '''
    A custom template for the X and Y Axis options in the right click context menu.
    This is a complete implementation of the original template.
    The purpose of these is to allow custom implementation of templates.
    Currently, it is a complete one-to-one copy of the original ones.
    '''
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(186, 154)
        Form.setMaximumSize(QSize(200, 16777215))
        self.gridLayout = QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setObjectName("gridLayout")
        self.label = QLabel(Form)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 7, 0, 1, 2)
        self.linkCombo = QComboBox(Form)
        self.linkCombo.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.linkCombo.setObjectName("linkCombo")
        self.gridLayout.addWidget(self.linkCombo, 7, 2, 1, 2)
        self.autoPercentSpin = QSpinBox(Form)
        self.autoPercentSpin.setEnabled(True)
        self.autoPercentSpin.setMinimum(1)
        self.autoPercentSpin.setMaximum(100)
        self.autoPercentSpin.setSingleStep(1)
        self.autoPercentSpin.setProperty("value", 100)
        self.autoPercentSpin.setObjectName("autoPercentSpin")
        self.gridLayout.addWidget(self.autoPercentSpin, 2, 2, 1, 2)
        self.autoRadio = QRadioButton(Form)
        self.autoRadio.setChecked(True)
        self.autoRadio.setObjectName("autoRadio")
        self.gridLayout.addWidget(self.autoRadio, 2, 0, 1, 2)
        self.manualRadio = QRadioButton(Form)
        self.manualRadio.setObjectName("manualRadio")
        self.gridLayout.addWidget(self.manualRadio, 1, 0, 1, 2)
        self.minText = QLineEdit(Form)
        self.minText.setObjectName("minText")
        self.gridLayout.addWidget(self.minText, 1, 2, 1, 1)
        self.maxText = QLineEdit(Form)
        self.maxText.setObjectName("maxText")
        self.gridLayout.addWidget(self.maxText, 1, 3, 1, 1)
        self.invertCheck = QCheckBox(Form)
        self.invertCheck.setObjectName("invertCheck")
        self.gridLayout.addWidget(self.invertCheck, 5, 0, 1, 4)
        self.mouseCheck = QCheckBox(Form)
        self.mouseCheck.setChecked(True)
        self.mouseCheck.setObjectName("mouseCheck")
        self.gridLayout.addWidget(self.mouseCheck, 6, 0, 1, 4)
        self.visibleOnlyCheck = QCheckBox(Form)
        self.visibleOnlyCheck.setObjectName("visibleOnlyCheck")
        self.gridLayout.addWidget(self.visibleOnlyCheck, 3, 2, 1, 2)
        self.autoPanCheck = QCheckBox(Form)
        self.autoPanCheck.setObjectName("autoPanCheck")
        self.gridLayout.addWidget(self.autoPanCheck, 4, 2, 1, 2)

        self.retranslateUi(Form)
        QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label.setText(_translate("Form", "Link Axis:"))
        self.linkCombo.setToolTip(_translate("Form", "<html><head/><body><p>Links this axis with another view. When linked, both views will display the same data range.</p></body></html>"))
        self.autoPercentSpin.setToolTip(_translate("Form", "<html><head/><body><p>Percent of data to be visible when auto-scaling. It may be useful to decrease this value for data with spiky noise.</p></body></html>"))
        self.autoPercentSpin.setSuffix(_translate("Form", "%"))
        self.autoRadio.setToolTip(_translate("Form", "<html><head/><body><p>Automatically resize this axis whenever the displayed data is changed.</p></body></html>"))
        self.autoRadio.setText(_translate("Form", "Auto"))
        self.manualRadio.setToolTip(_translate("Form", "<html><head/><body><p>Set the range for this axis manually. This disables automatic scaling. </p></body></html>"))
        self.manualRadio.setText(_translate("Form", "Manual"))
        self.minText.setToolTip(_translate("Form", "<html><head/><body><p>Minimum value to display for this axis.</p></body></html>"))
        self.minText.setText(_translate("Form", "0"))
        self.maxText.setToolTip(_translate("Form", "<html><head/><body><p>Maximum value to display for this axis.</p></body></html>"))
        self.maxText.setText(_translate("Form", "0"))
        self.invertCheck.setToolTip(_translate("Form", "<html><head/><body><p>Inverts the display of this axis. (+y points downward instead of upward)</p></body></html>"))
        self.invertCheck.setText(_translate("Form", "Invert Axis"))
        self.mouseCheck.setToolTip(_translate("Form", "<html><head/><body><p>Enables mouse interaction (panning, scaling) for this axis.</p></body></html>"))
        self.mouseCheck.setText(_translate("Form", "Mouse Enabled"))
        self.visibleOnlyCheck.setToolTip(_translate("Form", "<html><head/><body><p>When checked, the axis will only auto-scale to data that is visible along the orthogonal axis.</p></body></html>"))
        self.visibleOnlyCheck.setText(_translate("Form", "Visible Data Only"))
        self.autoPanCheck.setToolTip(_translate("Form", "<html><head/><body><p>When checked, the axis will automatically pan to center on the current data, but the scale along this axis will not change.</p></body></html>"))
        self.autoPanCheck.setText(_translate("Form", "Auto Pan Only"))


class ColorMapPlotWidget(InteractivePlotWidget):
    """An InteractivePlotWidget optimised for plotting color(heat) maps.
    Uses the Matplotlib colormap given by *cmap* to color the map.

    Attributes
    ----------
    lookup_table : ndarray
        The lookup table generated from *cmap* to colour the image with
    num_contours : int
        The number of different colour levels to plot
    contour_spacing : int
        How closely spaced the colour levels are
    """
    def __init__(self, parent=None, cmap="jet"):
        self.lookup_table = matplotlib_lookup_table(cmap)
        self.num_contours = 5
        self.contour_spacing_dB = 5
        self.parent = parent
        super().__init__(parent=self.parent)

    def plot_colormap(self, x, y, z, num_contours=5, contour_spacing_dB=5):
        """Plot *x*, *y* and *z* on a colourmap, with colour intervals defined
        by *num_contours* at *contour_spacing_dB* intervals."""

        #self.PlotWidget.removeItem(self.z_img)

        self.x = x
        self.y = y
        self.z = z

        self.num_contours = num_contours
        self.contour_spacing_dB = contour_spacing_dB
        self.update_lowest_contour()

        # Set up axes:
        x_axis = self.PlotWidget.getAxis('bottom')
        y_axis = self.PlotWidget.getAxis('left')

        self.x_scale_fact = self.get_scale_fact(x)
        self.y_scale_fact = self.get_scale_fact(y)

        x_axis.setScale(self.x_scale_fact)
        y_axis.setScale(self.y_scale_fact)

        #self.autoRange()

        self.z_img = ImageItem(z.transpose())
        self.z_img.setLookupTable(self.lookup_table)
        self.z_img.setLevels([self.lowest_contour, self.highest_contour])

        self.PlotWidget.addItem(self.z_img)

        self.PlotWidget.autoRange()
        #self.PlotWidget.ViewBox.autoRange()

    def get_scale_fact(self, var):
        return var.max() / var.size

    def update_lowest_contour(self):
        """Find the lowest contour to plot, as determined by the number of
        contours and the contour spacing."""
        self.lowest_contour = self.z.max() - (self.num_contours * self.contour_spacing_dB)
        self.highest_contour = self.z.max()


if __name__ == '__main__':

    pg.setConfigOption('background', 'w')
    pg.setConfigOption('foreground', 'k')
    pg.setConfigOption('antialias', True)
    defaultpen = pg.mkPen('k')

    app = 0
    app = QApplication(sys.argv)
    w = InteractivePlotWidget()
    x = np.linspace(0, 20*np.pi, 1e4)
    y = np.sin(x)
    w.PlotItem.plot(x, y, pen=defaultpen)
    w.PlotItem.autoRange()
    w.show()

    sys.exit(app.exec_())



"""
This module contains the class to record data from a National Instrument.
It uses PyDAQmx to do so, but requires NIDAQmx drivers to function.
Please check the PyDAQMx and NIDAQmx C API reference for more information.

Typical example of using the module:
>>>import myRecorder as NIR
>>>recorder = NIR.Recorder()
Channels: 1
Rate: 30000
Chunk size: 1000
Number of chunks: 4
You are using National Instrument for recording
Input device name not found, using the first device
Selected devices: Dev3
>>>recorder.stream_init()
Channels Name: Dev3/ai0
True
>>>recorder.record_init()
Recording function is ready! Use record_start() to start
True
>>>recorder.record_start()
stream already started
Recording Start!
True
>>>Recording Done! Please flush the data with flush_record_data().
data = recorder.flush_record_data()
Data flushed
>>>recorder.close()
"""
from cued_datalogger.acquisition.RecorderParent import RecorderParent

import sys,traceback
try:
    import PyDAQmx as pdaq
    from PyDAQmx import Task
except:
    # If PyDAQmx doesn't work, create mock version of it
    if sys.version_info >= (3,):
        from unittest.mock import MagicMock
    else:
        from mock import Mock as MagicMock
    class MockModule(MagicMock):
        @classmethod
        def __getattr__(cls, name):
                return MagicMock()

    sys.modules['PyDAQmx'] = MockModule()
    import PyDAQmx
    from PyDAQmx import Task

import numpy as np
import pprint as pp

class Recorder(RecorderParent):
    """
     Sets up the recording stream through a National Instrument

     Attributes
     ----------
    device_name: str
        Name of the device to be used for recording
    max_value: float
        Maximum value of recorded data
    """
#---------------- INITIALISATION METHODS -----------------------------------
    def __init__(self,channels = 1,rate = 30000.0, chunk_size = 1000,
                 num_chunk = 4,device_name = None):
        """
        Re-implemented from RecorderParent
        """

        super().__init__(channels = channels,rate = rate,
             chunk_size = chunk_size,num_chunk = num_chunk)
        print('You are using National Instrument for recording')

        self.device_name = None
        self.set_device_by_name(device_name);

        self.open_recorder()
        self.trigger_init()

        self.max_value = 10;

#---------------- DEVICE SETTING METHODS -----------------------------------
    def set_device_by_name(self, name):
        """
         Set the recording audio device by name.
         Uses the first device found if no such device found.
        """
        devices = self.available_devices()[0]
        selected_device = None
        if not devices:
            print('No NI devices found')
            return

        if not name in devices:
            print('Input device name not found, using the first device')
            selected_device = devices[0]
        else:
            selected_device = name

        print('Selected devices: %s' % selected_device)
        self.device_name = selected_device

     # Get audio device names
    def available_devices(self):
        """
        Get all the available input National Instrument devices.

        Returns
        ----------
        devices_name: List of str
            Name of the device, e.g. Dev0
        device_type: List of str
            Type of device, e.g. USB-6003
        """
        numBytesneeded = pdaq.DAQmxGetSysDevNames(None,0)
        databuffer = pdaq.create_string_buffer(numBytesneeded)
        pdaq.DAQmxGetSysDevNames(databuffer,numBytesneeded)

        #device_list = []
        devices_name = pdaq.string_at(databuffer).decode('utf-8').split(',')

        device_type = []
        for dev in devices_name:
            numBytesneeded = pdaq.DAQmxGetDevProductType(dev,None,0)
            databuffer = pdaq.create_string_buffer(numBytesneeded)
            pdaq.DAQmxGetDevProductType(dev,databuffer,numBytesneeded)
            device_type.append(pdaq.string_at(databuffer).decode('utf-8'))

        #device_list.append(devices_name)
        #device_list.append(device_type)

        return(devices_name,device_type)

    # Display the current selected device info
    def current_device_info(self):
        """
        Prints information about the current device set
        """
        device_info = {}
        info = ('Category', 'Type','Product', 'Number',
                'Analog Trigger Support','Analog Input Trigger Types','Analog Input Channels (ai)', 'Analog Output Channels (ao)',
                'ai Minimum Rate(Hz)', 'ai Maximum Rate(Single)(Hz)', 'ai Maximum Rate(Multi)(Hz)',
                'Digital Trigger Support','Digital Input Trigger Types','Digital Ports', 'Digital Lines', 'Terminals')
        funcs = (pdaq.DAQmxGetDevProductCategory, pdaq.DAQmxGetDevProductType,
                 pdaq.DAQmxGetDevProductNum, pdaq.DAQmxGetDevSerialNum,
                 pdaq.DAQmxGetDevAnlgTrigSupported,  pdaq.DAQmxGetDevAITrigUsage,
                 pdaq.DAQmxGetDevAIPhysicalChans,pdaq.DAQmxGetDevAOPhysicalChans,
                 pdaq.DAQmxGetDevAIMinRate, pdaq.DAQmxGetDevAIMaxSingleChanRate, pdaq.DAQmxGetDevAIMaxMultiChanRate,
                 pdaq.DAQmxGetDevDigTrigSupported,pdaq.DAQmxGetDevDITrigUsage,
                 pdaq.DAQmxGetDevDIPorts,pdaq.DAQmxGetDevDILines,
                 pdaq.DAQmxGetDevTerminals)
        var_types = (pdaq.int32, str, pdaq.uint32, pdaq.uint32,
                     pdaq.bool32,pdaq.int32,str,str,
                     pdaq.float64, pdaq.float64, pdaq.float64,
                     pdaq.bool32,pdaq.int32,str,str,str)

        for i,f,v in zip(info,funcs,var_types):
            try:
                if v == str:
                    nBytes = f(self.device_name,None,0)
                    string_ptr = pdaq.create_string_buffer(nBytes)
                    f(self.device_name,string_ptr,nBytes)
                    if any( x in i for x in ('Channels','Ports')):
                        device_info[i] = len(string_ptr.value.decode().split(','))
                    else:
                        device_info[i] = string_ptr.value.decode()
                else:
                    data = v()
                    f(self.device_name,data)
                    if 'Types' in i:
                        device_info[i] = bin(data.value)[2:].zfill(6)
                    else:
                        device_info[i] = data.value
            except Exception as e:
                print(e)
                device_info[i] = '-'

        pp.pprint(device_info)

    def set_channels(self):
        """
        Create the string to initiate the channels when assigning a Task

        Returns
        ----------
        channelname: str
            The channel names to be used when assigning Task
            e.g. Dev0/ai0:Dev0/ai1
        """
        if self.channels >1:
            channelname =  '%s/ai0:%s/ai%i' % (self.device_name, self.device_name,self.channels-1)
        elif self.channels == 1:
            channelname = '%s/ai0' % self.device_name

        print('Channels Name: %s' % channelname)
        return channelname

#---------------- STREAMING METHODS -----------------------------------
    # Convert data obtained into a proper array
    def audiodata_to_array(self,data):
        """
        Re-implemented from RecorderParent
        """
        return data.reshape((-1,self.channels))/(2**15) *10.0

    # Callback function for audio streaming
    def stream_audio_callback(self):
        """
        Callback function for audio streaming.
        First, it writes data to the circular buffer,
        then record data if it is recording,
        finally check for any trigger.

        Returns 0 as part of the callback format.
        More info can be found in PyDAQmx documentation on Task class
        """
        in_data = np.zeros(self.chunk_size*self.channels,dtype = np.int16)
        read = pdaq.int32()
        self.audio_stream.ReadBinaryI16(self.chunk_size,10.0,pdaq.DAQmx_Val_GroupByScanNumber,
                           in_data,self.chunk_size*self.channels,pdaq.byref(read),None)

        data_array = self.audiodata_to_array(in_data)
        self.write_buffer(data_array)
        #self.rEmitter.newdata.emit()

        if self.recording:
            self.record_data(data_array)
         # Trigger check
        if self.trigger:
            self._trigger_check_threshold(data_array)

        return 0

    def stream_init(self, playback = False):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream == None:
            try:
                self.audio_stream = Task()
                self.audio_stream.stream_audio_callback = self.stream_audio_callback
                self.audio_stream.CreateAIVoltageChan(self.set_channels(),"",
                                         pdaq.DAQmx_Val_RSE,-10.0,10.0,
                                         pdaq.DAQmx_Val_Volts,None)
                self.audio_stream.CfgSampClkTiming("",self.rate,
                                      pdaq.DAQmx_Val_Rising,pdaq.DAQmx_Val_ContSamps,
                                      self.chunk_size)
                self.audio_stream.AutoRegisterEveryNSamplesEvent(pdaq.DAQmx_Val_Acquired_Into_Buffer,
                                                    1000,0,name = 'stream_audio_callback')

                self.stream_start()
                return True
            except:
                t,v,tb = sys.exc_info()
                print(t)
                print(v)
                print(traceback.format_tb(tb))
                self.audio_stream = None

                return False

    # Start the streaming
    def stream_start(self):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream:
            task_done = pdaq.bool32()
            self.audio_stream.GetTaskComplete(task_done)
            if task_done.value:
                self.audio_stream.StartTask()
            else:
                print('stream already started')
        else:
            print('No audio stream is set up')
    # Stop the streaming
    def stream_stop(self):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream:
            task_done = pdaq.bool32()
            self.audio_stream.GetTaskComplete(task_done)
            if not task_done.value:
                self.audio_stream.StopTask()
            else:
                print('stream already stopped')
        else:
            print('No audio stream is set up')

    # Close the stream, probably needed if any parameter of the stream is changed
    def stream_close(self):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream:
            self.audio_stream.StopTask()
            self.audio_stream.ClearTask()
            self.audio_stream = None


import sys,traceback
import scipy.io as sio
from cued_datalogger.api.channel import Channel, DataSet, ChannelSet
import numpy as np
from PyQt5.QtWidgets import (QWidget, QVBoxLayout,QPushButton,QLabel,QTreeWidget,
                             QTreeWidgetItem,QHBoxLayout,QFileDialog)
from PyQt5.QtCore import  Qt
import pickle

def import_from_mat(file, channel_set=None):
    """
    A function for importing data and metadata to a ChannelSet from an 
    old-style DataLogger ``.mat`` file.
    
    Parameters
    ----------
    file : path_to_file
        The path to the ``.mat`` file to import data from.
    channel_set : ChannelSet
        The ChannelSet to save the imported data and metadata to. If ``None``, 
        a new ChannelSet is created and returned.
    """
    if channel_set is None:
        new_channel_set = True
        channel_set = ChannelSet()
    else:
        new_channel_set = False
    
    # Load the matlab file as a dict
    file = sio.loadmat(file)

    # Work out how many channels to create
    num_time_series_datasets = file["dt2"][0][0]
    num_spectrum_datasets = file["dt2"][0][1]
    num_sonogram_datasets = file["dt2"][0][2]

    num_channels = np.amax(np.asarray([num_time_series_datasets,
                                       num_spectrum_datasets,
                                       num_sonogram_datasets]))

    # # Extract metadata
    sample_rate = file["freq"][0][0]

    if "tsmax" in file.keys():
        time_series_scale_factor = file["tsmax"][0][0]
    else:
        time_series_scale_factor = 1

    if "npts" in file.keys():
        fft_num_samples = file["npts"]
    else:
        fft_num_samples = None

    if "tfun" in file.keys():
        is_transfer_function = bool(file["tfun"])
    else:
        is_transfer_function = False

    if "step" in file.keys():
        sonogram_fft_step = file["step"]
    else:
        sonogram_fft_step = None

    # # Extract data
    # Transpose the data so it's easier to work with:
    # In the matlab file it is in the form
    # (num_samples_per_channel, num_channels) - each channel is a column
    # Numpy works more easily if it is in the form
    # (num_channels, num_samples_per_channel) - each channel is a row
    if "indata" in file.keys():
        time_series = file["indata"].transpose()
    if "yspec" in file.keys():
        spectrum = file["yspec"].transpose()
    if "yson" in file.keys():
        sonogram_amplitude = file["yson"].transpose()
    if "yphase" in file.keys():
        sonogram_phase = file["yphase"].transpose()

    # # Save everything
    for i in np.arange(num_channels,dtype = np.int):
        # Create a new channel
        channel_set.add_channels()

        # Set channel metadata
        channel_set.set_channel_metadata(i, {"name": "Imported {}".format(i),
                                             "sample_rate": sample_rate,
                                             "calibration_factor":
                                                 time_series_scale_factor})

        # Set channel data
        if i < num_time_series_datasets:
            channel_set.add_channel_dataset(i,
                                            "time_series",
                                            time_series[i])

        # Differentiate between TF and FFT
        if i < num_spectrum_datasets:
            if is_transfer_function:
                channel_set.add_channel_dataset(i,
                                                "TF",
                                                 spectrum[i],
                                                 ' ')
            else:
                channel_set.add_channel_dataset(i,
                                                "spectrum",
                                                 spectrum[i],
                                                 'Hz')
            print(channel_set.channels[i].data("frequency"))
            
        if i < num_sonogram_datasets:
            channel_set.add_channel_dataset(i,
                                            "sonogram",
                                            sonogram_amplitude[i])
            channel_set.add_channel_dataset(i,
                                            "sonogram_phase",
                                            sonogram_phase[i],
                                            'rad')
    if new_channel_set:
        return channel_set
            
class DataImportWidget(QWidget):
    def __init__(self,parent):
        super().__init__(parent)
        self.new_cs = ChannelSet()
        self.init_UI()
        
    def init_UI(self):
        layout = QVBoxLayout(self)
        
        self.import_btn = QPushButton('Import Mat Files',self)
        self.import_btn.clicked.connect(self.import_files)
        self.pickle_btn = QPushButton('Import Pickle Files',self)
        self.pickle_btn.clicked.connect(self.load_pickle)
        layout.addWidget(self.import_btn)
        layout.addWidget(self.pickle_btn)
        layout.addWidget(QLabel('New ChannelSet Preview',self))
        
        self.tree = QTreeWidget(self)
        self.tree.setHeaderLabels(["Channel Number", "Name", "Units",
                                   "Comments", "Tags", "Sample rate",
                                   "Calibration factor",
                                   "Transfer function type"])
    
        layout.addWidget(self.tree)
        
        self.add_data_btn = QPushButton('Add ChannelSet to existing ChannelSet',self)
        self.rep_data_btn = QPushButton('Replace existing ChannelSet',self)
        layout.addWidget(self.add_data_btn)
        layout.addWidget(self.rep_data_btn)
        
    def set_channel_set(self, channel_set):
        print("Setting channel set...")
        self.tree.clear()

        #self.cs = channel_set

        self.channel_items = []

        for channel_number, channel in enumerate(channel_set.channels):
            # Create a tree widget item for this channel
            channel_item = QTreeWidgetItem(self.tree)
            #channel_item.setFlags(channel_item.flags() | Qt.ItemIsEditable)
            channel_item.setData(0, Qt.DisplayRole, channel_number)
            channel_item.setData(1, Qt.DisplayRole, channel.name)
            channel_item.setData(3, Qt.DisplayRole, channel.comments)
            channel_item.setData(4, Qt.DisplayRole, channel.tags)
            channel_item.setData(5, Qt.DisplayRole, '%.2f' % channel.sample_rate )
            channel_item.setData(6, Qt.DisplayRole, '%.2f' % channel.calibration_factor)
            channel_item.setData(7, Qt.DisplayRole, channel.transfer_function_type)
            # Add it to the list
            self.channel_items.append(channel_item)

            # Create a child tree widget item for each of the channel's datasets
            for dataset in channel.datasets:
                dataset_item = QTreeWidgetItem(channel_item)
                #dataset_item.setFlags(dataset_item.flags() | Qt.ItemIsEditable)
                dataset_item.setData(1, Qt.DisplayRole, dataset.id_)
                dataset_item.setData(2, Qt.DisplayRole, dataset.units)
        print("Done.")
        
    def import_files(self):
        # Get a list of URLs from a QFileDialog
        url = QFileDialog.getOpenFileNames(self, "Load transfer function", "addons",
                                               "MAT Files (*.mat)")[0]        
        try:
            import_from_mat(url[0], self.new_cs)
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            print('Load failed.')
            return
    
        self.set_channel_set(self.new_cs)
    
    def load_pickle(self):
        '''
        This is probably a temporary solution to loading data.
        Probably have to write a better way of storing data.
        PLEASE DO NOT OPEN ANY UNTRUSTED PICKLE FILES.
        UNPICKLING A FILE CAN EXECUTE ARBITRARY CODE, WHICH IS DANGEROUS TO YOUR COMPUTER.
        
        '''
        url = QFileDialog.getOpenFileName(self, "Load Channel Set", "addons",
                                               "Pickle Files (*.pickle)")[0]
        with open(url,'rb') as f:
            self.new_cs = pickle.load(f)
        self.set_channel_set(self.new_cs)
    
    def clear(self):
        self.new_cs = ChannelSet()
        self.set_channel_set(self.new_cs)
        
if __name__ == '__main__':
    cs = ChannelSet()
    import_from_mat("//cued-fs/users/general/tab53/ts-home/Documents/owncloud/Documents/urop/labs/4c6/transfer_function_clean.mat",
                    cs)
        
    print(cs.channel_ids(0))




if __name__ == '__main__':
    import sys
    sys.path.append('../')

import numpy as np
import pyqtgraph as pg

from cued_datalogger.api.numpy_extensions import MatlabList

from PyQt5.QtCore import pyqtSignal, Qt
from PyQt5.QtWidgets import (QWidget, QPushButton, QVBoxLayout,
                             QLineEdit, QCheckBox, QScrollArea,
                             QTreeWidget, QTreeWidgetItem, QHBoxLayout)

class ChannelSet(object):
    """
    A group of channels, with methods for setting and getting data.

    In theory, a user will only need to interact with the ChannelSet to
    interact with the channels and data. Each ChannelSet will normally be
    derived from one set of results, or one run of an experiment, and then
    the ChannelSet will contain all the information and analysis from that run.
    ChannelSets can be initialised as empty, and channels added later,
    or initialised with a number of empty channels, to which DataSets can be
    added later. Channels are stored in a matlab-style list structure
    (see :class:`MatlabList <cued_datalogger.api.numpy_extensions.MatlabList>`)
    which uses tuple indexing, eg. ``channelset.channels[1, 2, range(5,10)]``,
    so that multiple channels can be selected easily.


    Attributes
    ----------
    channels : MatlabList
        A list of the channels in this set.

    colormap : ColorMap
        A :class:`ColorMap` used for colouring the channels in this set.
    """
    def __init__(self, initial_num_channels=0):
        """Create the ChannelSet with a number of blank channels as given by
        *initial_num_channels*."""
        # Initialise the channel list
        self.channels = MatlabList()

        # Create an initial number of channels
        self.add_channels(initial_num_channels)

    def __len__(self):
        """Return the number of Channels in this ChannelSet."""
        return len(self.channels)

    def info(self):
        """Print the information about all of the Channels in this
        ChannelSet."""
        print("ChannelSet: \n Channels: \n")

        for i, channel in enumerate(self.channels):
            print("--------------------\n"
                + "{}    ".format(i))
            channel.info()

    def add_channels(self, num_channels=1):
        """Add a number (*num_channels*) of new empty Channels to the end of
        the channel list."""
        for i in range(num_channels):
            self.channels.append(Channel())

        self.update_channel_colours()

    def add_channel_dataset(self, channel_index, id_, data=None, units=None):
        """Add a DataSet with *id\_* to the Channel specified by
        *channel_index*. DataSet can be initialised as empty (default) or with
        *data* and/or *units*."""
        # If an tuple is given, indexing the channels will give an iterable,
        # otherwise it will give one result
        if isinstance(channel_index, tuple):
            for channel in self.channels[channel_index]:
                channel.add_dataset(id_)
        else:
            self.channels[channel_index].add_dataset(id_)

        if data is not None:
            self.set_channel_data(channel_index, id_, data)
        if units is not None:
            self.set_channel_units(channel_index, id_, units)

    def set_channel_data(self, channel_index, id_, data):
        """Set the data of DataSet with *id\_* to *data* in the Channel
        specified by *channel_index*."""
        # If an tuple is given, indexing the channels will give an iterable,
        # otherwise it will give one result
        if isinstance(channel_index, tuple):
            for channel in self.channels[channel_index]:
                channel.set_data(id_, data)
        else:
            self.channels[channel_index].set_data(id_, data)

    def set_channel_units(self, channel_index, id_, units):
        """Set the units of DataSet with *id\_* to *units* in the Channel
        specified by *channel_index*."""
        # If an tuple is given, indexing the channels will give an iterable,
        # otherwise it will give one result
        if isinstance(channel_index, tuple):
            for channel in self.channels[channel_index]:
                channel.set_units(id_, units)
        else:
            self.channels[channel_index].set_units(id_, units)

    def set_channel_metadata(self, channel_index, metadata_dict):
        """Set metadata of the Channel specified by *channel_index* using
        the keys and values given in *metadata_dict*."""
        # If an tuple is given, indexing the channels will give an iterable,
        # otherwise it will give one result
        if isinstance(channel_index, tuple):
            # Iterate through the given channels
            for channel in self.channels[channel_index]:
                # Get metadata from this channel
                channel.set_metadata(metadata_dict)
        else:
            # Get metadata from this channel
            self.channels[channel_index].set_metadata(metadata_dict)

    def update_channel_colours(self):
        """Update the :attr:`colormap` so that the channels are mapped to
        the full range of colours, and update all the channel colours."""
        # Set the PyQtGraph colormap to encompass the full range
        blue = [0, 0, 255, 255]
        red = [255, 0, 0, 255]
        green = [0, 255, 0, 255]
        colour_stops = [0, (len(self) - 1)//2, len(self) - 1]
        self.colormap = pg.ColorMap(colour_stops, [blue, green, red])

        for i in range(len(self)):
            self.set_channel_colour(i)

    def set_channel_colour(self, channel_index):
        """Set the RGBA tuple specifying the colour of the Channel at
        *channel_index* to a value determined by its index and the ChannelSet
        :attr:`colormap`."""
        self.channels[channel_index].colour = self.colormap.map(channel_index)

    def channel_ids(self, channel_index):
        """Return the ids of all the datasets in the Channel specified by
        *channel_index*."""
        # If an tuple is given, indexing the channels will give an iterable,
        # otherwise it will give one result
        if isinstance(channel_index, tuple):
            #output = []
            # Iterate through the given channels
            #for channel in self.channels[channel_index]:
                # Get metadata from this channel
            #    output.append(channel.ids())
            #return output
            return [channel.ids() for channel in
                    self.channels[channel_index]]
        else:
            # Get metadata from this channel
            return self.channels[channel_index].ids()

    def channel_data(self, channel_index, id_):
        """Return the data from the DataSet given by *id\_* in the Channel
        specified by *channel_index*."""
        # If an tuple is given, indexing the channels will give an iterable,
        # otherwise it will give one result
        if isinstance(channel_index, tuple):
            #for channel in self.channels[channel_index]:
            #    return channel.data(id_)
            return [channel.data(id_)
                    for channel in self.channels[channel_index]]
        else:
            return self.channels[channel_index].data(id_)

    def channel_units(self, channel_index, id_):
        """Return the units from the DataSet given by *id\_* in the Channel
        specified by *channel_index*."""
        # If an int is given, indexing the channels will give one result,
        # otherwise it will give an iterable
        if isinstance(channel_index, tuple):
            #for channel in self.channels[channel_index]:
            #    return channel.units(id_)
            return [channel.units(id_)
                    for channel in self.channels[channel_index]]
        else:
            return self.channels[channel_index].units(id_)

    def channel_metadata(self, channel_index, metadata_id=None):
        """Return the metadata (either a specific item given by *metadata_id*
        or the full dict of metadata) of the Channel specified by
        *channel_index*."""
        # If an int is given, indexing the channels will give one result,
        # otherwise it will give an iterable
        if isinstance(channel_index, tuple):
            # Iterate through the given channels
            #for channel in self.channels[channel_index]:
                # Get metadata from this channel
                #return channel.metadata(metadata_id)
            return [channel.metadata(metadata_id)
                    for channel in self.channels[channel_index]]
        else:
            # Get metadata from this channel
            return self.channels[channel_index].metadata(metadata_id)

    def channel_colour(self, channel_index):
        """Get the RGBA tuple specifying the colour of the Channel at
        *channel_index*."""
        return self.channels[channel_index].colour

class Channel(object):
    """
    Contains a group of DataSets and associated metadata.

    Channels are the basic structure used throughout the CUED DataLogger. Channels
    may contain many DataSets, but each must have a unique id\_ (ie. cannot
    have two 'time' DataSets). Typically a Channel will be initialised with
    just 'time_series' data, and other DataSets will be added as analysis is
    performed - eg. a Fourier Transform produces a 'spectrum' DataSet.
    Channels also contain metadata about the data.

    Attributes
    ----------
    name : str
        A human-readable string identifying this channel
        (eg. ``'Input 0'``, or ``'Left Accelerometer'``).

    datasets : list
        A list of this Channel's DataSets.

    comments : str
        A string for any additional comments.

    tags : list
        A list of tags (eg. ['accelerometer', 'input']) for quick
        selection and sorting.

    sample_rate : float
        The rate (in Hz) that the data was sampled at.

    calibration_factor : float
        #TODO#

    transfer_function_type : str
        Either 'None', 'displacement', 'velocity', or 'acceleration'-
        indicates what type of transfer function is stored.

    colour : tuple
        An RGBA tuple for this channel's colour - usually set
        by its parent ChannelSet
    """
    def __init__(self, name='', datasets=[],
                 comments='',
                 tags=[],
                 sample_rate=1000,
                 calibration_factor=1,
                 transfer_function_type="displacement",
                 colour=None):
        """Create a new Channel.
        Can be initialised as empty, or with given metadata and/or with given
        DataSets."""

        # Set the channel metadata
        self.name = name
        self.comments = comments
        self.tags = tags
        self.sample_rate = sample_rate
        self.calibration_factor = calibration_factor
        self.transfer_function_type = transfer_function_type
        self.datasets = []
        self.colour = colour

        # Create the auto-generated datasets
        self.add_dataset("time", 's')
        self.add_dataset("frequency", 'Hz')
        self.add_dataset("omega", 'rad')

        # Set the channel datasets
        for ds in datasets:
            self.add_dataset(ds.id_, ds.units, ds.data)

        self.update_autogenerated_datasets()

    def info(self):
        """Print this Channel's attributes, including DataSet ids
        and metadata."""
        print("""Name: {}
        DataSets: {}

        Metadata:
            sample_rate: {}
            calibration_factor: {}
            transfer_function_type: {}

            tags: {}
            comments: {}
        """.format(self.name,
                    self.ids(),
                    self.sample_rate,
                    self.calibration_factor,
                    self.transfer_function_type,
                    self.tags,
                    self.comments))

    def is_dataset(self, id_):
        """Return a boolean of whether the dataset given by *id\_*
        exists already."""
        return any([ds.id_ == id_ for ds in self.datasets])

    def add_dataset(self, id_, units=None, data=0):
        """Create a new dataset in this channel with *id\_*, *units*, *data*,
        unless a dataset given by *id\_* exists."""
        # Add as a new DataSet, unless it already exists
        if not self.is_dataset(id_):
            self.datasets.append(DataSet(id_, units, data))
            self.update_autogenerated_datasets()
        else:
            # If a dataset already exist, then it is fine
            print("DataSet '{}' already exists".format(id_))
            #raise ValueError("DataSet '{}' already exists".format(id_))

    def set_data(self, id_, data):
        """Set the data in dataset *id\_* to *data*."""
        # Set the data for a pre-existing DataSet
        for ds in self.datasets:
            if ds.id_ == id_:
                ds.set_data(data)
                self.update_autogenerated_datasets()
                return
        raise ValueError("No such DataSet '{}'".format(id_))

    def set_units(self, id_, units):
        """Set the units of dataset *id\_* to *units*."""
        # Set the units for a pre-existing DataSet
        for ds in self.datasets:
            if ds.id_ == id_:
                ds.set_units(units)
                return
        raise ValueError("No such DataSet '{}'".format(id_))

    def set_metadata(self, metadata_dict):
        """Set the channel metadata to the metadata given in
        *metadata_dict*."""
        for metadata_name, metadata_value in metadata_dict.items():
            # If a permitted item of metadata is given, set metadata
            if hasattr(self, metadata_name):
                setattr(self, metadata_name, metadata_value)
            else:
                raise ValueError("No such metadata '{}'".format(metadata_name))

    def dataset(self, id_):
        """Return the DataSet in this channel with *id\_*."""
        for ds in self.datasets:
            if ds.id_ == id_:
                return ds
        raise ValueError("No such DataSet {}".format(id_))

    def ids(self):
        """Return a list of the DataSet ids that this channel has."""
        # Get a list of the datasets that this channel has
        return [ds.id_ for ds in self.datasets]

    def data(self, id_):
        """Return the data from the DataSet given by *id\_*."""
        for ds in self.datasets:
            if ds.id_ == id_:
                return ds.data
        # If no dataset found, return an empty array instead
        print("No such DataSet {}".format(id_))
        return np.array([])
        #raise ValueError("No such DataSet {}".format(id_))

    def units(self, id_):
        """Return the units from the DataSet given by *id\_*."""
        for ds in self.datasets:
            if ds.id_ == id_:
                return ds.units
        raise ValueError("No such DataSet {}".format(id_))

    def metadata(self, metadata_id=None):
        """Return the value of this channel's metadata associated with
        *metadata_id*. If none given, returns all of this channel's
        metadata in a dictionary."""
        metadata_dict = {"name": self.name,
                         "comments": self.comments,
                         "tags": self.tags,
                         "sample_rate": self.sample_rate,
                         "calibration_factor": self.calibration_factor,
                         "transfer_function_type": self.transfer_function_type}
        if metadata_id is None:
            return metadata_dict
        else:
            for key, value in metadata_dict.items():
                    if key == metadata_id:
                        return value

    def update_autogenerated_datasets(self):
        """Regenerate the values in the automatically generated DataSets."""
        if self.is_dataset("time_series") or self.is_dataset("sonogram"):
            t = self.dataset("time")
            t.set_data(np.linspace(0, self.data("time_series").size / self.sample_rate, self.data("time_series").size))
        # Both TF and FFT requires frequency bins
        if self.is_dataset("spectrum") or self.is_dataset("TF") or self.is_dataset("sonogram"):
            freq = self.dataset("frequency")
            #f.set_data(np.linspace(0, self.sample_rate, self.data("spectrum").size))
            # Theo: sample_rate is divided by 2 because rfft return half of the actual FFT
            if self.is_dataset("spectrum"):
                dataset = self.data("spectrum")
            elif self.is_dataset("TF"):
                dataset = self.data("TF")
            elif self.is_dataset("sonogram"):
                dataset = self.data("sonogram")
            freq.set_data(np.linspace(0, self.sample_rate/2, dataset.size))
            w = self.dataset("omega")
            w.set_data(freq.data * 2*np.pi)


class DataSet(object):
    """
    A DataSet is the basic unit for data storage - a named 1d vector with units.

    Attributes
    ----------
    id\_ : str
        A lower-case string containing the name of the data stored
        in the vector. See Notes for permitted values.

    units : str
        The SI unit in which the data is measured.

    data : ndarray
        A numpy array of data points associated with id\_.

    Notes
    -----
    Permitted values for the DataSet :attr:`id\_` are:

    * ``"time_series"`` - The raw input time series data

    * ``"time"``\*	- Calculated from the sample rate and number of samples
      (units ``'s'``)

    * ``"frequency"``\* - Calculated from the sample rate and number of samples
      (units ``'Hz'``)

    * ``"omega"``\* - Angular frequency (units ``'rad'``), calculated from
      the sample rate and number of samples

    * ``"spectrum"`` - The complex spectrum given by the Fourier Transform

    * ``"sonogram"`` - The complex sonogram array, with shape (number of
      FFTs, frequencies)

    * ``"sonogram_frequency"``\* - The frequency bins (Hz) used in plotting the
      sonogram. Calculated from the sonogram parameters.

    * ``"sonogram_omega"``\* - The frequency bins (rad) used in plotting the
      sonogram. Calculated from the sonogram parameters.

    (\* indicates that this DataSet is auto-generated by the Channel)
    """
    def __init__(self, id_, units=None, data=np.array(0)):
        """Create a new DataSet with unique *id_*. Can either be initialised as
        empty, or with units and/or data."""
        self.set_id(id_)
        self.set_units(units)
        self.set_data(data)

    def set_id(self, id_):
        """Set the DataSet's id\_ to *id_*."""
        # Check that the user has input a permitted id_ type
        if isinstance(id_, str):
            self.id_ = id_
        else:
            raise TypeError("'id_' must be a string type.")

    def set_data(self, data):
        """Set the DataSet's data array to *data*."""
        # Set the dataset data
        self.data = np.asarray(data)

    def set_units(self, units):
        """Set the DataSet's units to *units*."""
        self.units = units


class ChannelSelectWidget(QWidget):
    """
    A widget used in the Global Toolbox to select channels.

    This widget is used as the master controller of what channels are selected.
    It allows channel selection by checkboxes, an 'Invert Selection' button,
    a 'Select All' button, a 'Deselect All' button, and Matlab-style list
    indexing (eg. ``1:10:2, 4`` selects all the odd channels between 1 and 10
    and channel 4). Possible additional features to be implemented include
    selection by tag and by other channel metadata.

    When the channel selection is changed it emits a signal containing the
    list of currently selected channels. Widgets can be set to receive this
    signal and set the channels that they are displaying to that list.

    Attributes
    ----------
    sig_channel_selection_changed : pyqtSignal
        The signal emitted when the selected channels are changed, containing
        a list of :class:`Channel` objects
    """

    sig_channel_selection_changed = pyqtSignal(list)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.init_ui()

    def init_ui(self):
        # Create the master layout
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        # # Create the viewbox of checkboxes
        # Create the viewbox
        self.checkbox_viewbox = QWidget(self)
        # Set the layout
        self.viewbox_layout = QVBoxLayout()
        self.checkbox_viewbox.setLayout(self.viewbox_layout)
        # Create a scroll area so that the box can be scrollable
        self.scrollbox = QScrollArea(self)
        self.scrollbox.setWidget(self.checkbox_viewbox)
        self.scrollbox.setWidgetResizable(True)
        # Add the viewbox to the master layout
        self.layout.addWidget(self.scrollbox)

        # # Create the selection toggle buttons
        self.selection_btns_layout = QVBoxLayout()
        self.layout.addLayout(self.selection_btns_layout)

        self.select_all_btn = QPushButton('Select All', self)
        self.select_all_btn.clicked.connect(self.select_all)
        self.select_all_btn.clicked.connect(self.on_channel_selection_change)
        self.selection_btns_layout.addWidget(self.select_all_btn)

        self.deselect_all_btn = QPushButton('Deselect All', self)
        self.deselect_all_btn.clicked.connect(self.deselect_all)
        self.deselect_all_btn.clicked.connect(self.on_channel_selection_change)
        self.selection_btns_layout.addWidget(self.deselect_all_btn)

        self.invert_select_btn = QPushButton('Invert Selection', self)
        self.invert_select_btn.clicked.connect(self.invert_select)
        self.invert_select_btn.clicked.connect(self.on_channel_selection_change)
        self.selection_btns_layout.addWidget(self.invert_select_btn)

        # # Create the text selection box
        self.text_select_box = QLineEdit(self)
        self.text_select_box.returnPressed.connect(self.select_by_text)
        self.selection_btns_layout.addWidget(self.text_select_box)

    def select_all(self):
        self.text_select_box.clear()

        for checkbox in self.checkbox_list:
            checkbox.setChecked(True)

    def deselect_all(self):
       self.text_select_box.clear()

       for checkbox in self.checkbox_list:
            checkbox.setChecked(False)

    def invert_select(self):
        self.text_select_box.clear()

        for checkbox in self.checkbox_list:
            checkbox.toggle()

    def select_by_text(self):
        # Get the text from the box
        string = self.text_select_box.text()
        #print("Selecting by " + string)

        selected_list = []

        # Split the string by commas
        index_list = string.split(",")
        for index in index_list:
            # If it's just a number, add it to the list
            if index.isdigit():
                selected_list.append(int(index))

            # If it's a slice, add the sliced bits to the list
            split_on_colon = index.split(":")
            if len(split_on_colon) > 1:
                # If it's a slice with no step
                if len(split_on_colon) == 2:
                    for i in range(int(split_on_colon[0]),
                                   int(split_on_colon[1]) + 1):
                        selected_list.append(i)
                # If it's a slice with step
                if len(split_on_colon) == 3:
                    for i in range(int(split_on_colon[0]),
                                   int(split_on_colon[1]) + 1,
                                   int(split_on_colon[2])):
                        selected_list.append(i)
            else:
                # Ignore anything else
                continue

        self.deselect_all()

        self.set_selected(selected_list)

    def set_selected(self, list_to_select):
        for channel_num, channel in enumerate(self.cs.channels):
            if channel_num in list_to_select:
                self.checkbox_list[channel_num].setChecked(True)

        self.on_channel_selection_change()

    def on_channel_selection_change(self):
        # Emit the "Selection changed" signal with a list of channels
        # that are currently selected
        print("Currently selected channels: {}".format(self.selected_channels_index()))
        self.sig_channel_selection_changed.emit(self.selected_channels())

    def set_channel_name(self):
        for i, channel in enumerate(self.cs.channels):
            # Create a checkbox for this channel
            self.checkbox_list[i].setText("{}: {}".format(i, channel.name))

    def set_channel_set(self, channel_set):
        """Set the :class:`ChannelSet` used by this widget."""
        self.cs = channel_set

        self.checkbox_list = []

        # Clear the old layout
        while self.viewbox_layout.count():
            self.viewbox_layout.takeAt(0).widget().deleteLater()

        for i, channel in enumerate(self.cs.channels):
            # Create a checkbox for this channel
            checkbox = QCheckBox("{}: {}".format(i, channel.name),
                                 self.checkbox_viewbox)
            checkbox.setChecked(True)
            checkbox.clicked.connect(self.on_channel_selection_change)
            # Add it to the list
            self.checkbox_list.append(checkbox)
            # Add it to the layout
            self.viewbox_layout.addWidget(self.checkbox_list[i])

        # Send out a signal with the updated channels
        self.on_channel_selection_change()

    def selected_channels_index(self):
        """Return a list of channel numbers of all currently selected
        channels."""
        selected_list = []

        for i, channel in enumerate(self.cs.channels):
            if self.checkbox_list[i].isChecked():
                selected_list.append(i)

        return selected_list

    def selected_channels(self):
        """Return a list of all the currently selected :class:`Channel`
        objects."""
        selected_list = []

        for i, channel in enumerate(self.cs.channels):
            if self.checkbox_list[i].isChecked():
                selected_list.append(channel)

        return selected_list


class ChannelMetadataWidget(QWidget):
    metadataChange = pyqtSignal(ChannelSet)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.init_ui()

    def init_ui(self):
        # Create the master layout
        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        # # Create the tree widget
        self.tree = QTreeWidget()
        self.tree.setHeaderLabels(["Channel Number", "Name", "Units",
                                   "Comments", "Tags", "Sample rate",
                                   "Calibration factor",
                                   "Transfer function type"])
        # Connect the signals
        self.tree.itemDoubleClicked.connect(self.edit_item)
        # Add it to the layout
        self.layout.addWidget(self.tree)

        # # Create the buttons
        self.button_layout = QHBoxLayout()
        self.layout.addLayout(self.button_layout)

        self.discard_button = QPushButton("Discard changes", self)
        self.discard_button.clicked.connect(self.discard_changes)
        self.button_layout.addWidget(self.discard_button)

        self.save_button = QPushButton("Save changes", self)
        self.save_button.clicked.connect(self.update_channelset)
        self.button_layout.addWidget(self.save_button)

    def set_channel_set(self, channel_set):
        print("Setting channel set...")
        self.tree.clear()

        self.cs = channel_set

        self.channel_items = []

        for channel_number, channel in enumerate(self.cs.channels):
            # Create a tree widget item for this channel
            channel_item = QTreeWidgetItem(self.tree)
            #channel_item.setFlags(channel_item.flags() | Qt.ItemIsEditable)
            channel_item.setData(0, Qt.DisplayRole, channel_number)
            channel_item.setData(1, Qt.DisplayRole, channel.name)
            channel_item.setData(3, Qt.DisplayRole, channel.comments)
            channel_item.setData(4, Qt.DisplayRole, channel.tags)
            channel_item.setData(5, Qt.DisplayRole, '%.2f' % channel.sample_rate)
            channel_item.setData(6, Qt.DisplayRole, '%.2f' % channel.calibration_factor)
            channel_item.setData(7, Qt.DisplayRole, channel.transfer_function_type)
            # Add it to the list
            self.channel_items.append(channel_item)

            # Create a child tree widget item for each of the channel's datasets
            for dataset in channel.datasets:
                dataset_item = QTreeWidgetItem(channel_item)
                #dataset_item.setFlags(dataset_item.flags() | Qt.ItemIsEditable)
                dataset_item.setData(1, Qt.DisplayRole, dataset.id_)
                dataset_item.setData(2, Qt.DisplayRole, dataset.units)
        print("Done.")

    def update_channelset(self):
        print("Updating channelset from tree...")

        for channel_number, channel_item in enumerate(self.channel_items):
            # Reset data that should be fixed (channel numbers)
            channel_item.setData(0, Qt.DisplayRole, channel_number)

            # Get the metadata from the tree
            # Note: data acquired from tree is string, need to convert some to float/int
            name = channel_item.data(1, Qt.DisplayRole)
            comments = channel_item.data(3, Qt.DisplayRole)
            tags = channel_item.data(4, Qt.DisplayRole)
            sample_rate = float(channel_item.data(5, Qt.DisplayRole))
            calibration_factor = float(channel_item.data(6, Qt.DisplayRole))
            transfer_function_type = channel_item.data(7, Qt.DisplayRole)

            # Set the channel metadata
            metadata_dict = {"name": name,
                             "comments": comments,
                             "tags": tags,
                             "sample_rate": sample_rate,
                             "calibration_factor": calibration_factor,
                             "transfer_function_type": transfer_function_type}
            self.cs.set_channel_metadata(channel_number, metadata_dict)

            # Update the datasets as well
            for i in range(channel_item.childCount()):
                # Find the dataset item
                dataset_item = channel_item.child(i)
                # Extract the data from the tree
                id_ = dataset_item.data(1, Qt.DisplayRole)
                units = dataset_item.data(2, Qt.DisplayRole)
                # Set the dataset units
                self.cs.set_channel_units(channel_number, id_, units)

        self.metadataChange.emit(self.cs)
        print("Done.")

    def edit_item(self, item, column):
        print("Editing item, column {}...".format(column))
        # Channel numbers are non-editable
        if column == 0:
            print("Channel numbers are non-editable")
            pass
        # Dataset ids are a non-editable
        elif item.parent() is not None and column == 1:
            print("Dataset ids are non-editable")
            pass
        else:
            # Set item to be editable
            old_flags = item.flags()
            item.setFlags(item.flags() | Qt.ItemIsEditable)
            # Edit the item
            self.tree.editItem(item, column)
            # Set item to not be editable
            item.setFlags(old_flags)
        print("Done.")

    def discard_changes(self):
        self.set_channel_set(self.cs)

if __name__ == '__main__':
    from api.file_import import import_from_mat
    from PyQt5.QtWidgets import QApplication
    app = 0
    app = QApplication(sys.argv)

    cs = ChannelSet()
    import_from_mat("//cued-fs/users/general/tab53/ts-home/Documents/owncloud/Documents/urop/labs/4c6/transfer_function_grid.mat", cs)

    #w = ChannelSelectWidget()
    w = ChannelMetadataWidget()
    w.set_channel_set(cs)

    w.show()

    sys.exit(app.exec_())

